from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import sys as _sys
import mdp
from mdp import Node, NodeException, numx, numx_rand
from mdp.nodes import WhiteningNode
from mdp.utils import (DelayCovarianceMatrix, MultipleCovarianceMatrices,
                       rotate, mult)


# TODO: support floats of size different than 64-bit; will need to change SQRT_EPS_D

# rename often used functions
sum, cos, sin, PI = numx.sum, numx.cos, numx.sin, numx.pi
SQRT_EPS_D = numx.sqrt(numx.finfo('d').eps)

def _triu(m, k=0):
    """ returns the elements on and above the k-th diagonal of m.  k=0 is the
        main diagonal, k > 0 is above and k < 0 is below the main diagonal."""
    N = m.shape[0]
    M = m.shape[1]
    x = numx.greater_equal(numx.subtract.outer(numx.arange(N),
                                               numx.arange(M)),1-k)
    out = (1-x)*m
    return out

#############
class ISFANode(Node):
    """
    Perform Independent Slow Feature Analysis on the input data.

    **Internal variables of interest**

      ``self.RP``
          The global rotation-permutation matrix. This is the filter
          applied on input_data to get output_data

      ``self.RPC``
          The *complete* global rotation-permutation matrix. This
          is a matrix of dimension input_dim x input_dim (the 'outer space'
          is retained)

      ``self.covs``
          A `mdp.utils.MultipleCovarianceMatrices` instance containing
          the current time-delayed covariance matrices of the input_data.
          After convergence the uppermost ``output_dim`` x ``output_dim``
          submatrices should be almost diagonal.
          
          ``self.covs[n-1]`` is the covariance matrix relative to the ``n``-th
          time-lag
            
          Note: they are not cleared after convergence. If you need to free
          some memory, you can safely delete them with::
          
              >>> del self.covs

      ``self.initial_contrast``
          A dictionary with the starting contrast and the SFA and ICA parts of
          it.

      ``self.final_contrast``
          Like the above but after convergence.

    Note: If you intend to use this node for large datasets please have
    a look at the ``stop_training`` method documentation for
    speeding things up.

    References:
    Blaschke, T. , Zito, T., and Wiskott, L. (2007).
    Independent Slow Feature Analysis and Nonlinear Blind Source Separation.
    Neural Computation 19(4):994-1021 (2007)
    http://itb.biologie.hu-berlin.de/~wiskott/Publications/BlasZitoWisk2007-ISFA-NeurComp.pdf
    """
    def __init__(self, lags=1, sfa_ica_coeff=(1., 1.), icaweights=None,
                 sfaweights=None, whitened=False, white_comp = None,
                 white_parm = None, eps_contrast=1e-6, max_iter=10000,
                 RP=None, verbose=False, input_dim=None, output_dim=None,
                 dtype=None):
        """
        Perform Independent Slow Feature Analysis.

        The notation is the same used in the paper by Blaschke et al. Please
        refer to the paper for more information.

        :Parameters:
          lags
            list of time-lags to generate the time-delayed covariance
            matrices (in the paper this is the set of \tau). If
            lags is an integer, time-lags 1,2,...,'lags' are used.
            Note that time-lag == 0 (instantaneous correlation) is
            always implicitly used.

          sfa_ica_coeff
            a list of float with two entries, which defines the
            weights of the SFA and ICA part of the objective
            function. They are called b_{SFA} and b_{ICA} in the
            paper.

          sfaweights
            weighting factors for the covariance matrices relative
            to the SFA part of the objective function (called
            \kappa_{SFA}^{\tau} in the paper). Default is
            [1., 0., ..., 0.]
            For possible values see the description of icaweights.

          icaweights
            weighting factors for the cov matrices relative
            to the ICA part of the objective function (called
            \kappa_{ICA}^{\tau} in the paper). Default is 1.
            Possible values are:

            - an integer ``n``: all matrices are weighted the same
              (note that it does not make sense to have ``n != 1``)

            - a list or array of floats of ``len == len(lags)``:
              each element of the list is used for weighting the
              corresponding matrix

            - ``None``: use the default values.

          whitened
            ``True`` if input data is already white, ``False``
            otherwise (the data will be whitened internally).

          white_comp
            If whitened is false, you can set ``white_comp`` to the
            number of whitened components to keep during the
            calculation (i.e., the input dimensions are reduced to
            ``white_comp`` by keeping the components of largest variance).
          white_parm
            a dictionary with additional parameters for whitening.
            It is passed directly to the WhiteningNode constructor.
            Ex: white_parm = { 'svd' : True }

          eps_contrast
            Convergence is achieved when the relative
            improvement in the contrast is below this threshold.
            Values in the range [1E-4, 1E-10] are usually
            reasonable.

          max_iter
            If the algorithms does not achieve convergence within
            max_iter iterations raise an Exception. Should be
            larger than 100.

          RP
            Starting rotation-permutation matrix. It is an
            input_dim x input_dim matrix used to initially rotate the
            input components. If not set, the identity matrix is used.
            In the paper this is used to start the algorithm at the
            SFA solution (which is often quite near to the optimum).

          verbose
            print progress information during convergence. This can
            slow down the algorithm, but it's the only way to see
            the rate of improvement and immediately spot if something
            is going wrong.

          output_dim
            sets the number of independent components that have to
            be extracted. Note that if this is not smaller than
            input_dim, the problem is solved linearly and SFA
            would give the same solution only much faster.
        """
        # check that the "lags" argument has some meaningful value
        if isinstance(lags, (int, int)):
            lags = list(range(1, lags+1))
        elif isinstance(lags, (list, tuple)):
            lags = numx.array(lags, "i")
        elif isinstance(lags, numx.ndarray):
            if not (lags.dtype.char in ['i', 'l']):
                err_str = "lags must be integer!"
                raise NodeException(err_str)
            else:
                pass
        else:
            err_str = ("Lags must be int, list or array. Found "
                       "%s!" % (type(lags).__name__))
            raise NodeException(err_str)
        self.lags = lags

        # sanity checks for weights
        if icaweights is None:
            self.icaweights = 1.
        else:
            if (len(icaweights) != len(lags)):
                err = ("icaweights vector length is %d, "
                       "should be %d" % (str(len(icaweights)), str(len(lags))))
                raise NodeException(err)
            self.icaweights = icaweights

        if sfaweights is None:
            self.sfaweights = [0]*len(lags)
            self.sfaweights[0] = 1.
        else:
            if (len(sfaweights) != len(lags)):
                err = ("sfaweights vector length is %d, "
                       "should be %d" % (str(len(sfaweights)), str(len(lags))))
                raise NodeException(err)
            self.sfaweights = sfaweights

        # store attributes
        self.sfa_ica_coeff = sfa_ica_coeff
        self.max_iter = max_iter
        self.verbose = verbose
        self.eps_contrast = eps_contrast

        # if input is not white, insert a WhiteningNode
        self.whitened = whitened
        if not whitened:
            if white_parm is None:
                white_parm = {}
            if output_dim is not None:
                white_comp = output_dim
            elif white_comp is not None:
                output_dim = white_comp
            self.white = WhiteningNode(input_dim=input_dim,
                                       output_dim=white_comp,
                                       dtype=dtype, **white_parm)

        # initialize covariance matrices
        self.covs = [ DelayCovarianceMatrix(dt, dtype=dtype) for dt in lags ]

        # initialize the global rotation-permutation matrix
        # if not set that we'll eventually be an identity matrix
        self.RP = RP

        # initialize verbose structure to print nice and useful progress info
        if verbose:
            info = { 'sweep' : max(len(str(self.max_iter)), 5),
                     'perturbe': max(len(str(self.max_iter)), 5),
                     'float' : 5+8,
                     'fmt' : "%.5e",
                     'sep' : " | "}
            f1 = "Sweep".center(info['sweep'])
            f1_2 = "Pertb". center(info['perturbe'])
            f2 = "SFA part".center(info['float'])
            f3 = "ICA part".center(info['float'])
            f4 = "Contrast".center(info['float'])
            header = info['sep'].join([f1, f1_2, f2, f3, f4])
            info['header'] = header+'\n'
            info['line'] = len(header)*"-"
            self._info = info

        # finally call base class constructor
        super(ISFANode, self).__init__(input_dim, output_dim, dtype)

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node.
      
        Support floating point types with size larger or equal than 64 bits.
        """
        return [t for t in mdp.utils.get_dtypes('Float') if t.itemsize>=8]

    def _set_dtype(self, dtype):
        # when typecode is set, we set the whitening node if needed and
        # the SFA and ICA weights
        self._dtype = dtype
        if not self.whitened and self.white.dtype is None:
            self.white.dtype = dtype
        self.icaweights = numx.array(self.icaweights, dtype)
        self.sfaweights = numx.array(self.sfaweights, dtype)

    def _set_input_dim(self, n):
        self._input_dim = n
        if not self.whitened and self.white.output_dim is not None:
            self._effective_input_dim = self.white.output_dim
        else:
            self._effective_input_dim = n

    def _train(self, x):
        # train the whitening node if needed
        if not self.whitened:
            self.white.train(x)
        # update the covariance matrices
        [self.covs[i].update(x) for i in range(len(self.lags))]

    def _execute(self, x):
        # filter through whitening node if needed
        if not self.whitened:
            x = self.white.execute(x)
        # rotate input
        return mult(x, self.RP)

    def _inverse(self, y):
        # counter-rotate input
        x = mult(y, self.RP.T)
        # invert whitening node if needed
        if not self.whitened:
            x = self.white.inverse(x)
        return x

    def _fmt_prog_info(self, sweep, pert, contrast, sfa = None, ica = None):
        # for internal use only!
        # format the progress information
        # don't try to understand this code: it Just Works (TM)
        fmt = self._info
        sweep_str = str(sweep).rjust(fmt['sweep'])
        pert_str = str(pert).rjust(fmt['perturbe'])
        if sfa is None:
            sfa_str = fmt['float']*' '
        else:
            sfa_str = (fmt['fmt']%(sfa)).rjust(fmt['float'])
        if ica is None:
            ica_str = fmt['float']*' '
        else:
            ica_str = (fmt['fmt'] % (ica)).rjust(fmt['float'])
        contrast_str = (fmt['fmt'] % (contrast)).rjust(fmt['float'])
        table_entry = fmt['sep'].join([sweep_str,
                                       pert_str,
                                       sfa_str,
                                       ica_str,
                                       contrast_str])
        return table_entry

    def _get_eye(self):
        # return an identity matrix with the right dimensions and type
        return numx.eye(self._effective_input_dim, dtype=self.dtype)

    def _get_rnd_rotation(self, dim):
        # return a random rot matrix with the right dimensions and type
        return mdp.utils.random_rot(dim, self.dtype)

    def _get_rnd_permutation(self, dim):
        # return a random permut matrix with the right dimensions and type
        zero = numx.zeros((dim, dim), dtype=self.dtype)
        row = numx_rand.permutation(dim)
        for col in range(dim):
            zero[row[col], col] = 1.
        return zero

    def _givens_angle(self, i, j, covs, bica_bsfa=None, complete=0):
        # Return the Givens rotation angle for which the contrast function
        # is minimal
        if bica_bsfa is None:
            bica_bsfa = self._bica_bsfa
        if j < self.output_dim:
            return self._givens_angle_case1(i, j, covs,
                                            bica_bsfa, complete=complete)
        else:
            return self._givens_angle_case2(i, j, covs,
                                            bica_bsfa, complete=complete)


    def _givens_angle_case2(self, m, n, covs, bica_bsfa, complete=0):
        # This function makes use of the constants computed in the paper
        #
        # R -> R
        # m -> \mu
        # n -> \nu
        #
        # Note that the minus sign before the angle phi is there because
        # in the paper the rotation convention is the opposite of ours.

        ncovs = covs.ncovs
        covs = covs.covs
        icaweights = self.icaweights
        sfaweights = self.sfaweights
        R = self.output_dim
        bica, bsfa = bica_bsfa

        Cmm, Cmn, Cnn = covs[m, m, :], covs[m, n, :], covs[n, n, :]
        d0 =   (sfaweights * Cmm*Cmm).sum()
        d1 = 4*(sfaweights * Cmn*Cmm).sum()
        d2 = 2*(sfaweights * (2*Cmn*Cmn + Cmm*Cnn)).sum()
        d3 = 4*(sfaweights * Cmn*Cnn).sum()
        d4 =   (sfaweights * Cnn*Cnn).sum()
        e0 = 2*(icaweights * ((covs[:R, m, :]*covs[:R, m, :]).sum(axis=0)
                              - Cmm*Cmm)).sum()
        e1 = 4*(icaweights * ((covs[:R, m, :]*covs[:R, n, :]).sum(axis=0)
                              - Cmm*Cmn)).sum()
        e2 = 2*(icaweights * ((covs[:R, n, :]*covs[:R, n, :]).sum(axis=0)
                              - Cmn*Cmn)).sum()

        s22 = 0.25 * bsfa*(d1+d3)   + 0.5* bica*(e1)
        c22 = 0.5  * bsfa*(d0-d4)   + 0.5* bica*(e0-e2)
        s24 = 0.125* bsfa*(d1-d3)
        c24 = 0.125* bsfa*(d0-d2+d4)

        # Compute the contrast function in a grid of angles to find a
        # first approximation for the minimum.  Repeat two times
        # (effectively doubling the resolution). Note that we can do
        # that because we know we have a single minimum.
        #
        # npoints should not be too large otherwise the contrast
        # funtion appears to be constant. This is because we hit the
        # maximum resolution for the cosine function (ca. 1e-15)
        npoints = 100
        left = old_div(-PI,2) - old_div(PI,(npoints+1))
        right = old_div(PI,2) + old_div(PI,(npoints+1))
        for iter in (1, 2):
            phi = numx.linspace(left, right, npoints+3)
            contrast = c22*cos(-2*phi)+s22*sin(-2*phi)+\
                       c24*cos(-4*phi)+s24*sin(-4*phi)
            minidx = contrast.argmin()
            left = phi[max(minidx-1, 0)]
            right = phi[min(minidx+1, len(phi)-1)]

        # The contrast is almost a parabola around the minimum.
        # To find the minimum we can therefore compute the derivative
        # (which should be a line) and calculate its root.
        # This step helps to overcome the resolution limit of the
        # cosine function and clearly improve the final result.
        der_left = 2*c22*sin(-2*left)- 2*s22*cos(-2*left)+\
                   4*c24*sin(-4*left)- 4*s24*cos(-4*left)
        der_right = 2*c22*sin(-2*right)-2*s22*cos(-2*right)+\
                    4*c24*sin(-4*right)-4*s24*cos(-4*right)
        if abs(der_left - der_right) < SQRT_EPS_D:
            minimum = phi[minidx]
        else:
            minimum = right - der_right*(right-left)/(der_right-der_left)

        dc = numx.zeros((ncovs,), dtype = self.dtype)
        for t in range(ncovs):
            dg = covs[:R, :R, t].diagonal()
            dc[t] = (dg*dg).sum(axis=0)
        dc = ((dc-Cmm*Cmm)*sfaweights).sum()

        ec = numx.zeros((ncovs, ), dtype = self.dtype)
        for t in range(ncovs):
            ec[t] = sum([covs[i, j, t]*covs[i, j, t] for i in range(R-1)
                         for j in range(i+1, R) if i != m and j != m])
        ec = 2*(ec*icaweights).sum()
        a20 = 0.125*bsfa*(3*d0+d2+3*d4+8*dc)+0.5*bica*(e0+e2+2*ec)
        minimum_contrast = a20+c22*cos(-2*minimum)+s22*sin(-2*minimum)+\
                           c24*cos(-4*minimum)+s24*sin(-4*minimum)
        if complete:
            # Compute the contrast between -pi/2 and pi/2
            # (useful for testing purposes)
            npoints = 1000
            phi = numx.linspace(old_div(-PI,2), old_div(PI,2), npoints+1)
            contrast = a20 + c22*cos(-2*phi) + s22*sin(-2*phi) +\
                       c24*cos(-4*phi) + s24*sin(-4*phi)
            return phi, contrast, minimum, minimum_contrast
        else:
            return minimum, minimum_contrast


    def _givens_angle_case1(self, m, n, covs, bica_bsfa, complete=0):
        # This function makes use of the constants computed in the paper
        #
        # R -> R
        # m -> \mu
        # n -> \nu
        #
        # Note that the minus sign before the angle phi is there because
        # in the paper the rotation convention is the opposite of ours.
        ncovs = covs.ncovs
        covs = covs.covs
        icaweights = self.icaweights
        sfaweights = self.sfaweights
        bica, bsfa = bica_bsfa

        Cmm, Cmn, Cnn = covs[m, m, :], covs[m, n, :], covs[n, n, :]
        d0 =   (sfaweights * (Cmm*Cmm+Cnn*Cnn)).sum()
        d1 = 4*(sfaweights * (Cmm*Cmn-Cmn*Cnn)).sum()
        d2 = 2*(sfaweights * (2*Cmn*Cmn+Cmm*Cnn)).sum()
        e0 = 2*(icaweights * Cmn*Cmn).sum()
        e1 = 4*(icaweights * (Cmn*Cnn-Cmm*Cmn)).sum()
        e2 =   (icaweights * ((Cmm-Cnn)*(Cmm-Cnn)-2*Cmn*Cmn)).sum()

        s24 = 0.25* (bsfa * d1    + bica * e1)
        c24 = 0.25* (bsfa *(d0-d2)+ bica *(e0-e2))

        # compute the exact minimum
        # Note that 'arctan' finds always the first maximum
        # because s24sin(4p)+c24cos(4p)=const*cos(4p-arctan)
        # the minimum lies +pi/4 apart (period = pi/2).
        # In other words we want that: abs(minimum) < pi/4
        phi4 = numx.arctan2(s24, c24)
        # use if-structure until bug in numx.sign is solved
        if  phi4 >= 0:
            minimum = -0.25*(phi4-PI)
        else:
            minimum = -0.25*(phi4+PI)

        # compute all constants:
        R = self.output_dim
        dc = numx.zeros((ncovs, ), dtype = self.dtype)
        for t in range(ncovs):
            dg = covs[:R, :R, t].diagonal()
            dc[t] = (dg*dg).sum(axis=0)
        dc = ((dc-Cnn*Cnn-Cmm*Cmm)*sfaweights).sum()
        ec = numx.zeros((ncovs, ), dtype = self.dtype)
        for t in range(ncovs):
            triu_covs = _triu(covs[:R, :R, t], 1).ravel()
            ec[t] = ((triu_covs*triu_covs).sum() - covs[m, n, t]*covs[m, n, t])
        ec = 2*(icaweights*ec).sum()
        a20 = 0.25*(bsfa*(4*dc+d2+3*d0)+bica*(4*ec+e2+3*e0))
        minimum_contrast = a20+c24*cos(-4*minimum)+s24*sin(-4*minimum)
        npoints = 1000
        if complete == 1:
            # Compute the contrast between -pi/2 and pi/2
            # (useful for testing purposes)
            phi = numx.linspace(old_div(-PI,2), old_div(PI,2), npoints+1)
            contrast = a20 + c24*cos(-4*phi) + s24*sin(-4*phi)
            return phi, contrast, minimum, minimum_contrast
        elif complete == 2:
            phi = numx.linspace(old_div(-PI,4), old_div(PI,4), npoints+1)
            contrast = a20 + c24*cos(-4*phi) + s24*sin(-4*phi)
            return phi, contrast, minimum, minimum_contrast
        else:
            return minimum, minimum_contrast


    def _get_contrast(self, covs, bica_bsfa = None):
        if bica_bsfa is None:
            bica_bsfa = self._bica_bsfa
        # return current value of the contrast
        R = self.output_dim
        ncovs = covs.ncovs
        covs = covs.covs
        icaweights = self.icaweights
        sfaweights = self.sfaweights
        # unpack the bsfa and bica coefficients
        bica, bsfa = bica_bsfa
        sfa = numx.zeros((ncovs, ), dtype=self.dtype)
        ica = numx.zeros((ncovs, ), dtype=self.dtype)
        for t in range(ncovs):
            sq_corr =  covs[:R, :R, t]*covs[:R, :R, t]
            sfa[t] = sq_corr.trace()
            ica[t] = 2*_triu(sq_corr, 1).ravel().sum()
        return (bsfa*sfaweights*sfa).sum(), (bica*icaweights*ica).sum()

    def _adjust_ica_sfa_coeff(self):
        # adjust sfa/ica ratio. ica and sfa term are scaled
        # differently because sfa accounts for the diagonal terms
        # whereas ica accounts for the off-diagonal terms
        ncomp = self.output_dim
        if ncomp > 1:
            bica =  old_div(self.sfa_ica_coeff[1],(ncomp*(ncomp-1)))
            bsfa = old_div(-self.sfa_ica_coeff[0],ncomp)
        else:
            bica =  0.#self.sfa_ica_coeff[1]
            bsfa = -self.sfa_ica_coeff[0]
        self._bica_bsfa = [bica, bsfa]

    def _fix_covs(self, covs=None):
        # fiv covariance matrices
        if covs is None:
            covs = self.covs
            if not self.whitened:
                white = self.white
                white.stop_training()
                proj = white.get_projmatrix(transposed=0)
            else:
                proj = None
            # fix and whiten the covariance matrices
            for i in range(len(self.lags)):
                covs[i], avg, avg_dt, tlen = covs[i].fix(proj)

            # send the matrices to the container class
            covs = MultipleCovarianceMatrices(covs)
            # symmetrize the cov matrices
            covs.symmetrize()
        self.covs = covs

    def _optimize(self):
        # optimize contrast function

        # save initial contrast
        sfa, ica = self._get_contrast(self.covs)
        self.initial_contrast = {'SFA': sfa,
                                 'ICA': ica,
                                 'TOT': sfa + ica}
        # info headers
        if self.verbose:
            print(self._info['header']+self._info['line'])

        # initialize control variables
        # contrast
        contrast = sfa+ica
        # local rotation matrix
        Q = self._get_eye()
        # local copy of correlation matrices
        covs = self.covs.copy()
        # maximum improvement in the contrast function
        max_increase = self.eps_contrast
        # Number of sweeps
        sweep = 0
        # flag for stopping sweeping
        sweeping = True
        # flag to check if we already perturbed the outer space
        # - negative means that we exit from this routine
        #   because we hit numerical precision or because
        #   there's no outer space to be perturbed (input_dim == outpu_dim)
        # - positive means the number of perturbations done
        #   before finding no further improvement
        perturbed = 0

        # size of the perturbation matrix
        psize = self._effective_input_dim-self.output_dim

        # if there is no outer space don't perturbe
        if self._effective_input_dim == self.output_dim:
            perturbed = -1

        # local eye matrix
        eye = self._get_eye()

        # main loop
        # we'll keep on sweeping until the contrast has improved less
        # then self.eps_contrast
        part_sweep = 0
        while sweeping:
            # update number of sweeps
            sweep += 1

            # perform a single sweep
            max_increase, covs, Q, contrast = self._do_sweep(covs, Q, contrast)

            if max_increase < 0 or contrast == 0:
                # we hit numerical precision, exit!
                sweeping = False
                if perturbed == 0:
                    perturbed = -1
                else:
                    perturbed = -perturbed

            if (max_increase < self.eps_contrast) and (max_increase) >= 0 :
                # rate of change is small for all pairs in a sweep
                if perturbed == 0:
                    # perturbe the outer space one time with a random rotation
                    perturbed = 1
                elif perturbed >= 1 and part_sweep == sweep-1:
                    # after the last pertubation no useful step has
                    # been done. exit!
                    sweeping = False
                elif perturbed < 0:
                    # we can't perturbe anymore
                    sweeping = False
                # keep track of the last sweep we perturbed
                part_sweep = sweep

            # perform perturbation if needed
            if perturbed >= 1 and sweeping is True:
                # generate a random rotation matrix for the external subspace
                PRT = eye.copy()
                rot = self._get_rnd_rotation(psize)
                # generate a random permutation matrix for the ext. subspace
                perm = self._get_rnd_permutation(psize)
                # combine rotation and permutation
                rot_perm = mult(rot, perm)
                # apply rotation+permutation
                PRT[self.output_dim:, self.output_dim:] = rot_perm
                covs.transform(PRT)
                Q = mult(Q, PRT)
                # increment perturbation counter
                perturbed += 1

            # verbose progress information
            if self.verbose:
                table_entry = self._fmt_prog_info(sweep,
                                                  perturbed,
                                                  contrast)
                _sys.stdout.write(table_entry+len(table_entry)*'\b')
                _sys.stdout.flush()

            # if we made too many sweeps exit with error!
            if sweep == self.max_iter:
                err_str = ("Failed to converge, maximum increase= "
                           "%.5e" % (max_increase))
                raise NodeException(err_str)

        # if we land here, we have converged!
        # calculate output contrast

        sfa, ica =  self._get_contrast(covs)
        contrast = sfa+ica
        # print final information
        if self.verbose:
            print(self._fmt_prog_info(sweep, perturbed,
                                      contrast, sfa, ica))
            print(self._info['line'])

        self.final_contrast = {'SFA': sfa,
                               'ICA': ica,
                               'TOT': sfa + ica}

        # finally return optimal rotation matrix
        return Q

    def _do_sweep(self, covs, Q, prev_contrast):
        # perform a single sweep

        # initialize maximal improvement in a single sweep
        max_increase = -1
        # shuffle rotation order
        numx_rand.shuffle(self.rot_axis)
        # sweep through all axes combinations
        for (i, j) in self.rot_axis:
            # get the angle that minimizes the contrast
            # and the contrast value
            angle, contrast = self._givens_angle(i, j, covs)
            if contrast == 0:
                # we hit numerical precision in case when b_sfa == 0
                # we can only break things from here on, better quit!
                max_increase = -1
                break

            # relative improvement in the contrast function
            relative_diff = old_div((prev_contrast-contrast),abs(prev_contrast))
            if relative_diff < 0:
                # if rate of change is negative we hit numerical precision
                # or we already sit on the optimum for this pair of axis.
                # don't rotate anymore and go to the next pair
                continue

            # update the rotation matrix
            rotate(Q, angle, [i, j])
            # rotate the covariance matrices
            covs.rotate(angle, [i, j])

            # store maximum and previous rate of change
            max_increase = max(max_increase, relative_diff)
            prev_contrast = contrast

        return max_increase, covs, Q, contrast

    def _stop_training(self, covs=None):
        """Stop the training phase.

        If the node is used on large datasets it may be wise to first
        learn the covariance matrices, and then tune the parameters
        until a suitable parameter set has been found (learning the
        covariance matrices is the slowest part in this case).  This
        could be done for example in the following way (assuming the
        data is already white):

        >>> covs=[mdp.utils.DelayCovarianceMatrix(dt, dtype=dtype)
        ...       for dt in lags]
        >>> for block in data:
        ...     [covs[i].update(block) for i in range(len(lags))]

        You can then initialize the ISFANode with the desired parameters,
        do a fake training with some random data to set the internal
        node structure and then call stop_training with the stored covariance
        matrices. For example:

        >>> isfa = ISFANode(lags, .....)
        >>> x = mdp.numx_rand.random((100, input_dim)).astype(dtype)
        >>> isfa.train(x)
        >>> isfa.stop_training(covs=covs)

        This trick has been used in the paper to apply ISFA to surrogate
        matrices, i.e. covariance matrices that were not learnt on a
        real dataset.
        """
        # fix, whiten, symmetrize and weight the covariance matrices
        # the functions sets also the number of input components self.ncomp
        self._fix_covs(covs)
        # if output_dim were not set, set it to be the number of input comps
        if self.output_dim is None:
            self.output_dim = self._effective_input_dim
        # adjust b_sfa and b_ica
        self._adjust_ica_sfa_coeff()
        # initialize all possible rotation axes
        self.rot_axis = [(i, j) for i in range(0, self.output_dim)
                         for j in range(i+1, self._effective_input_dim)]

        # initialize the global rotation-permutation matrix (RP):
        RP = self.RP
        if RP is None:
            RP = self._get_eye()
        else:
            # apply the global rotation matrix
            self.covs.transform(RP)

        # find optimal rotation
        Q = self._optimize()
        RP = mult(RP, Q)
        # rotate and permute the covariance matrices
        # we do it here in one step, to avoid the cumulative errors
        # of multiple rotations in _optimize
        self.covs.transform(Q)

        # keep the complete rotation-permutation matrix
        self.RPC = RP.copy()

        # Reduce dimension to match output_dim#
        RP = RP[:, :self.output_dim]
        # the variance for the derivative of a whitened signal is
        # 0 <= v <= 4, therefore the diagonal elements of the delayed
        # covariance matrice with time lag = 1 (covs[:,:,0]) are
        # -1 <= v' <= +1
        # reorder the components to have them ordered by slowness
        d = (self.covs.covs[:self.output_dim, :self.output_dim, 0]).diagonal()
        idx = d.argsort()[::-1]
        self.RP = RP.take(idx, axis=1)

        # we could in principle clean up self.covs, as we do in SFANode or
        # PCANode, but this algorithm is not stable enough to rule out
        # possible problems. When these occcurs examining the covariance
        # matrices is often the only way to debug.
        #del self.covs



import sys
import mdp
import mdp.hinet as hinet
n = mdp.numx

from bimdp import BiNode


class BiSwitchboard(BiNode, hinet.Switchboard):
    """BiMDP version of the normal Switchboard.

    It adds support for stop_message and also tries to apply the switchboard
    mapping to arrays in the message. The mapping is only applied if the
    array is at least two dimensional and the second dimension matches the
    switchboard dimension.
    """

    def __init__(self, **kwargs):
        """Initialize BiSwitchboard.

        args and kwargs are forwarded via super to the next __init__ method
        in the MRO.
        """
        super(BiSwitchboard, self).__init__(**kwargs)
        if self.inverse_connections is not None:
            self.down_connections = self.inverse_connections
        else:
            # a stable (order preserving where possible) sort is
            # necessary here, therefore use mergesort
            # otherwise channels (e.g. for a Rectangular2d...) are mixed up
            self.down_connections = n.argsort(self.connections,
                                              kind="mergesort")

    def _execute(self, x, msg=None):
        """Return the routed input data."""
        if x is not None:
            y = super(BiSwitchboard, self)._execute(x)
        else:
            y = None
        msg = self._execute_msg(msg)
        if not msg:
            return y
        else:
            return y, msg

    def _inverse(self, x, msg=None):
        """Return the routed input data."""
        if x is not None:
            y = super(BiSwitchboard, self)._inverse(x)
        else:
            y = None
        msg = self._inverse_msg(msg)
        if not msg:
            return y
        else:
            return y, msg

    def is_bi_training(self):
        return False

    ## Helper methods ##

    def _inverse_msg(self, msg):
        """Inverse routing for msg."""
        if not msg:
            return None
        out_msg = {}
        for (key, value) in list(msg.items()):
            if (type(value) is n.ndarray and
                len(value.shape) >= 2 and value.shape[1] == self.output_dim):
                out_msg[key] = super(BiSwitchboard, self)._inverse(value)
            else:
                out_msg[key] = value
        return out_msg

    def _execute_msg(self, msg):
        """Feed-forward routing for msg."""
        if not msg:
            return None
        out_msg = {}
        for (key, value) in list(msg.items()):
            if (type(value) is n.ndarray and
                len(value.shape) >= 2 and value.shape[1] == self.input_dim):
                out_msg[key] = super(BiSwitchboard, self)._execute(value)
            else:
                out_msg[key] = value
        return out_msg


## create BiSwitchboard versions of the standard MDP switchboards ##

# corresponding methods for the switchboard_factory extension are
# created as well

@classmethod
def _binode_create_switchboard(cls, free_params, prev_switchboard,
                               prev_output_dim, node_id):
    """Modified version of create_switchboard to support node_id.

    This method can be used as a substitute when using the switchboard_factory
    extension.
    """
    compatible = False
    for base_class in cls.compatible_pre_switchboards:
        if isinstance(prev_switchboard, base_class):
            compatible = True
    if not compatible:
        err = ("The prev_switchboard class '%s'" %
                    prev_switchboard.__class__.__name__ +
               " is not compatible with this switchboard class" +
               " '%s'." % cls.__name__)
        raise mdp.hinet.SwitchboardException(err)
    for key, value in list(free_params.items()):
        if key.endswith('_xy') and isinstance(value, int):
            free_params[key] = (value, value)
    kwargs = cls._get_switchboard_kwargs(free_params, prev_switchboard,
                                         prev_output_dim)
    return cls(node_id=node_id, **kwargs)


# TODO: Use same technique as for binodes?
#    But have to take care of the switchboard_factory extension.

# use a function to avoid poluting the namespace
def _create_bi_switchboards():
    switchboard_classes = [
        mdp.hinet.ChannelSwitchboard,
        mdp.hinet.Rectangular2dSwitchboard,
        mdp.hinet.DoubleRect2dSwitchboard,
        mdp.hinet.DoubleRhomb2dSwitchboard,
    ]
    current_module = sys.modules[__name__]
    for switchboard_class in switchboard_classes:
        node_name = switchboard_class.__name__
        binode_name = node_name[:-len("Switchboard")] + "BiSwitchboard"
        docstring = ("Automatically created BiSwitchboard version of %s." %
                     node_name)
        docstring = "Automatically created BiNode version of %s." % node_name
        exec(('class %s(BiSwitchboard, mdp.hinet.%s): "%s"' %
              (binode_name, node_name, docstring)), current_module.__dict__)
        # create appropriate FactoryExtension nodes
        mdp.extension_method("switchboard_factory",
                             current_module.__dict__[binode_name],
                             "create_switchboard")(_binode_create_switchboard)

_create_bi_switchboards()



"""These are test functions for MDP utilities.
"""
from __future__ import division
from builtins import range
from past.utils import old_div
from builtins import object
import py.test
from ._tools import *
from mdp import Node, nodes

class BogusClass(object):
    def __init__(self):
        self.x = numx_rand.random((2,2))

class BogusNode(Node):
    x = numx_rand.random((2,2))
    y = BogusClass()
    z = BogusClass()
    z.z = BogusClass()


def test_introspection():
    bogus = BogusNode()
    arrays, string = utils.dig_node(bogus)
    assert len(list(arrays.keys())) == 4, 'Not all arrays where caught'
    assert sorted(arrays.keys()) == ['x', 'y.x',
                                     'z.x', 'z.z.x'], 'Wrong names'
    sizes = [x[0] for x in list(arrays.values())]
    assert sorted(sizes) == [numx_rand.random((2,2)).itemsize*4]*4, \
           'Wrong sizes'
    sfa = nodes.SFANode()
    sfa.train(numx_rand.random((1000, 10)))
    a_sfa, string = utils.dig_node(sfa)
    keys = ['_cov_mtx._avg', '_cov_mtx._cov_mtx',
            '_dcov_mtx._avg', '_dcov_mtx._cov_mtx']
    assert sorted(a_sfa.keys()) == keys, 'Wrong arrays in SFANode'
    sfa.stop_training()
    a_sfa, string = utils.dig_node(sfa)
    keys = ['_bias', 'avg', 'd', 'davg', 'sf']
    assert sorted(a_sfa.keys()) == keys, 'Wrong arrays in SFANode'

def test_random_rot():
    dim = 20
    tlen = 10
    for i in range(tlen):
        x = utils.random_rot(dim, dtype='f')
        assert x.dtype.char=='f', 'Wrong dtype'
        y = utils.mult(x.T, x)
        assert_almost_equal(numx_linalg.det(x), 1., 4)
        assert_array_almost_equal(y, numx.eye(dim), 4)

def test_random_rot_determinant_sign():
    x = utils.random_rot(4)
    assert_almost_equal(numx_linalg.det(x), 1., 4)
    x = utils.random_rot(5)
    assert_almost_equal(numx_linalg.det(x), 1., 4)


def test_casting():
    x = numx_rand.random((5,3)).astype('d')
    y = 3*x
    assert_type_equal(y.dtype, x.dtype)
    x = numx_rand.random((5,3)).astype('f')
    y = 3.*x
    assert_type_equal(y.dtype, x.dtype)
    x = (10*numx_rand.random((5,3))).astype('i')
    y = 3.*x
    assert_type_equal(y.dtype, 'd')
    y = 3*x
    assert_type_equal(y.dtype, 'i')
    x = numx_rand.random((5,3)).astype('f')
    y = 3*x
    assert_type_equal(y.dtype, 'f')

def test_mult_diag():
    dim = 20
    d = numx_rand.random(size=(dim,))
    dd = numx.diag(d)
    mtx = numx_rand.random(size=(dim, dim))

    res1 = utils.mult(dd, mtx)
    res2 = utils.mult_diag(d, mtx, left=True)
    assert_array_almost_equal(res1, res2, 10)
    res1 = utils.mult(mtx, dd)
    res2 = utils.mult_diag(d, mtx, left=False)
    assert_array_almost_equal(res1, res2, 10)

def test_symeig_fake_integer():
    a = numx.array([[1,2],[2,7]])
    b = numx.array([[3,1],[1,5]])
    w,z = utils._symeig._symeig_fake(a)
    w,z = utils._symeig._symeig_fake(a,b)

def test_symeig_fake_LAPACK_bug():
    # bug. when input matrix is almost an identity matrix
    # but not exactly, the lapack dgeev routine returns a
    # matrix of eigenvectors which is not orthogonal.
    # this bug was present when we used numx_linalg.eig
    # instead of numx_linalg.eigh .
    # Note: this is a LAPACK bug.
    y = numx_rand.random((4,4))*1E-16
    y = old_div((y+y.T),2)
    for i in range(4):
        y[i,i]=1
    val, vec = utils._symeig._symeig_fake(y)
    assert_almost_equal(abs(numx_linalg.det(vec)), 1., 12)

def test_QuadraticForm_extrema():
    # TODO: add some real test
    # check H with negligible linear term
    noise = 1e-8
    tol = 1e-6
    x = numx_rand.random((10,))
    H = numx.outer(x, x) + numx.eye(10)*0.1
    f = noise*numx_rand.random((10,))
    q = utils.QuadraticForm(H, f)
    xmax, xmin = q.get_extrema(utils.norm2(x), tol=tol)
    assert_array_almost_equal(x, xmax, 5)
    # check I + linear term
    H = numx.eye(10, dtype='d')
    f = x
    q = utils.QuadraticForm(H, f=f)
    xmax, xmin = q.get_extrema(utils.norm2(x), tol=tol)
    assert_array_almost_equal(f, xmax, 5)

def test_QuadraticForm_invariances():
    #nu = numx.linspace(2.,-3,10)
    nu = numx.linspace(6., 1, 10)
    H = utils.symrand(nu)
    E, W = mdp.utils.symeig(H)
    q = utils.QuadraticForm(H)
    xmax, xmin = q.get_extrema(5.)
    e_w, e_sd = q.get_invariances(xmax)
    #print e_sd,nu[1:]-nu[0]
    assert_array_almost_equal(e_sd,nu[1:]-nu[0],6)
    assert_array_almost_equal(abs(e_w),abs(W[:,-2::-1]),6)
    e_w, e_sd = q.get_invariances(xmin)
    assert_array_almost_equal(e_sd,nu[-2::-1]-nu[-1],6)
    assert_array_almost_equal(abs(e_w),abs(W[:,1:]),6)

def test_QuadraticForm_non_symmetric_raises():
    """Test the detection of non symmetric H!
    """
    H = numx_rand.random((10,10))
    py.test.raises(mdp.utils.QuadraticFormException,
                   utils.QuadraticForm, H)

def test_nongeneral_svd_bug():
    a = numx.array([[ 0.73083003,  0.        ,  0.7641788 ,  0.        ],
                    [ 0.        ,  0.        ,  0.        ,  0.        ],
                    [ 0.7641788 ,  0.        ,  0.79904932,  0.        ],
                    [ 0.        ,  0.        ,  0.        ,  0.        ]])
    w, z = utils.nongeneral_svd(a)
    diag = numx.diagonal(utils.mult(utils.hermitian(z),
                                    utils.mult(a, z))).real
    assert_array_almost_equal(diag, w, 12)



"""
Module for MDP classifiers that support parallel training.
"""

import mdp
from mdp import numx
from mdp.parallel import ParallelExtensionNode


class ParallelGaussianClassifier(ParallelExtensionNode,
                                 mdp.nodes.GaussianClassifier):

    def _fork(self):
        return self._default_fork()

    def _join(self, forked_node):
        if not self._cov_objs:
            self.set_dtype(forked_node._dtype)
            self._cov_objs = forked_node._cov_objs
        else:
            for key, forked_cov in list(forked_node._cov_objs.items()):
                if key in self._cov_objs:
                    self._join_covariance(self._cov_objs[key], forked_cov)
                else:
                    self._cov_objs[key] = forked_cov
                    

class ParallelNearestMeanClassifier(ParallelExtensionNode,
                                    mdp.nodes.NearestMeanClassifier):
    
    def _fork(self):
        return self._default_fork()
    
    def _join(self, forked_node):
        for key in forked_node.label_means:
            if key in self.label_means:
                self.label_means[key] += forked_node.label_means[key]
                self.n_label_samples[key] += forked_node.n_label_samples[key]
            else:
                self.label_means[key] = forked_node.label_means[key]
                self.n_label_samples[key] = forked_node.n_label_samples[key]


class ParallelKNNClassifier(ParallelExtensionNode,
                            mdp.nodes.KNNClassifier):
    
    def _fork(self):
        return self._default_fork()
    
    def _join(self, forked_node):
        for key in forked_node._label_samples:
            if key in self._label_samples:
                self._label_samples[key] += forked_node._label_samples[key]
            else:
                self._label_samples[key] = forked_node._label_samples[key]





from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from mdp import numx
from mdp.utils import (mult, nongeneral_svd, CovarianceMatrix,
                       symeig, SymeigException)
import warnings as _warnings

class PCANode(mdp.Node):
    """Filter the input data through the most significatives of its
    principal components.

    **Internal variables of interest**
    
      ``self.avg``
          Mean of the input data (available after training).

      ``self.v``
          Transposed of the projection matrix (available after training).

      ``self.d``
          Variance corresponding to the PCA components (eigenvalues of the
          covariance matrix).

      ``self.explained_variance``
          When output_dim has been specified as a fraction of the total
          variance, this is the fraction of the total variance that is
          actually explained.

    More information about Principal Component Analysis, a.k.a. discrete
    Karhunen-Loeve transform can be found among others in
    I.T. Jolliffe, Principal Component Analysis, Springer-Verlag (1986).
    """

    def __init__(self, input_dim=None, output_dim=None, dtype=None,
                 svd=False, reduce=False, var_rel=1E-12, var_abs=1E-15,
                 var_part=None):
        """The number of principal components to be kept can be specified as
        'output_dim' directly (e.g. 'output_dim=10' means 10 components
        are kept) or by the fraction of variance to be explained
        (e.g. 'output_dim=0.95' means that as many components as necessary
        will be kept in order to explain 95% of the input variance).

        Other Keyword Arguments:

        svd -- if True use Singular Value Decomposition instead of the
               standard eigenvalue problem solver. Use it when PCANode
               complains about singular covariance matrices

        reduce -- Keep only those principal components which have a variance
                  larger than 'var_abs' and a variance relative to the
                  first principal component larger than 'var_rel' and a
                  variance relative to total variance larger than 'var_part'
                  (set var_part to None or 0 for no filtering).
                  Note: when the 'reduce' switch is enabled, the actual number
                  of principal components (self.output_dim) may be different
                  from that set when creating the instance.
        """
        # this must occur *before* calling super!
        self.desired_variance = None
        super(PCANode, self).__init__(input_dim, output_dim, dtype)
        self.svd = svd
        # set routine for eigenproblem
        if svd:
            self._symeig = nongeneral_svd
        else:
            self._symeig = symeig
        self.var_abs = var_abs
        self.var_rel = var_rel
        self.var_part = var_part
        self.reduce = reduce
        # empirical covariance matrix, updated during the training phase
        self._cov_mtx = CovarianceMatrix(dtype)
        # attributes that defined in stop_training
        self.d = None  # eigenvalues
        self.v = None  # eigenvectors, first index for coordinates
        self.total_variance = None
        self.tlen = None
        self.avg = None
        self.explained_variance = None

    def _set_output_dim(self, n):
        if n <= 1 and isinstance(n, float):
            # set the output dim after training, when the variances are known
            self.desired_variance = n
        else:
            self._output_dim = n

    def _check_output(self, y):
        # check output rank
        if not y.ndim == 2:
            error_str = "y has rank %d, should be 2" % (y.ndim)
            raise mdp.NodeException(error_str)

        if y.shape[1] == 0 or y.shape[1] > self.output_dim:
            error_str = ("y has dimension %d"
                         ", should be 0<y<=%d" % (y.shape[1], self.output_dim))
            raise mdp.NodeException(error_str)

    def get_explained_variance(self):
        """Return the fraction of the original variance that can be
        explained by self._output_dim PCA components.
        If for example output_dim has been set to 0.95, the explained
        variance could be something like 0.958...
        Note that if output_dim was explicitly set to be a fixed number
        of components, there is no way to calculate the explained variance.
        """
        return self.explained_variance

    def _train(self, x):
        # update the covariance matrix
        self._cov_mtx.update(x)

    def _adjust_output_dim(self):
        """Return the eigenvector range and set the output dim if required.

        This is used if the output dimensions is smaller than the input
        dimension (so only the larger eigenvectors have to be kept).
        """
        # if the number of principal components to keep is not specified,
        # keep all components
        if self.desired_variance is None and self.output_dim is None:
            self.output_dim = self.input_dim
            return None

        ## define the range of eigenvalues to compute
        # if the number of principal components to keep has been
        # specified directly
        if self.output_dim is not None and self.output_dim >= 1:
            # (eigenvalues sorted in ascending order)
            return (self.input_dim - self.output_dim + 1,
                   self.input_dim)
        # otherwise, the number of principal components to keep has been
        # specified by the fraction of variance to be explained
        else:
            return None

    def _stop_training(self, debug=False):
        """Stop the training phase.

        Keyword arguments:

        debug=True     if stop_training fails because of singular cov
                       matrices, the singular matrices itselves are stored in
                       self.cov_mtx and self.dcov_mtx to be examined.
        """
        # request the covariance matrix and clean up
        self.cov_mtx, avg, self.tlen = self._cov_mtx.fix()
        del self._cov_mtx

        # this is a bit counterintuitive, as it reshapes the average vector to
        # be a matrix. in this way, however, we spare the reshape
        # operation every time that 'execute' is called.
        self.avg = avg.reshape(1, avg.shape[0])

        # range for the eigenvalues
        rng = self._adjust_output_dim()

        # if we have more variables then observations we are bound to fail here
        # suggest to use the NIPALSNode instead.
        if debug and self.tlen < self.input_dim:
            wrn = ('The number of observations (%d) '
                   'is larger than the number of input variables '
                   '(%d). You may want to use '
                   'the NIPALSNode instead.' % (self.tlen, self.input_dim))
            _warnings.warn(wrn, mdp.MDPWarning)

        # total variance can be computed at this point:
        # note that vartot == d.sum()
        vartot = numx.diag(self.cov_mtx).sum()

        ## compute and sort the eigenvalues
        # compute the eigenvectors of the covariance matrix (inplace)
        # (eigenvalues sorted in ascending order)
        try:
            d, v = self._symeig(self.cov_mtx, range=rng, overwrite=(not debug))
            # if reduce=False and svd=False. we should check for
            # negative eigenvalues and fail
            if not (self.reduce or self.svd or (self.desired_variance is
                                                not None)):
                if d.min() < 0:
                    raise mdp.NodeException(
                        "Got negative eigenvalues: %s.\n"
                        "You may either set output_dim to be smaller, "
                        "or set reduce=True and/or svd=True" % str(d))
        except SymeigException as exception:
            err = str(exception)+("\nCovariance matrix may be singular."
                                  "Try setting svd=True.")
            raise mdp.NodeException(err)

        # delete covariance matrix if no exception occurred
        if not debug:
            del self.cov_mtx

        # sort by descending order
        d = numx.take(d, list(range(d.shape[0]-1, -1, -1)))
        v = v[:, ::-1]

        if self.desired_variance is not None:
            # throw away immediately negative eigenvalues
            d = d[ d > 0 ]
            # the number of principal components to keep has
            # been specified by the fraction of variance to be explained
            varcum = (old_div(d, vartot)).cumsum(axis=0)
            # select only the relevant eigenvalues
            # number of relevant eigenvalues
            neigval = int(varcum.searchsorted(self.desired_variance) + 1.)
            #self.explained_variance = varcum[neigval-1]
            # cut
            d = d[0:neigval]
            v = v[:, 0:neigval]
            # define the new output dimension
            self.output_dim = int(neigval)

        # automatic dimensionality reduction
        if self.reduce:
            # remove entries that are smaller then var_abs and
            # smaller then var_rel relative to the maximum
            d = d[ d > self.var_abs ]
            # check that we did not throw away everything
            if len(d) == 0:
                raise mdp.NodeException('No eigenvalues larger than'
                                        ' var_abs=%e!'%self.var_abs)
            d = d[ old_div(d, d.max()) > self.var_rel ]

            # filter for variance relative to total variance
            if self.var_part:
                d = d[ old_div(d, vartot) > self.var_part ]

            v = v[:, 0:d.shape[0]]
            self._output_dim = d.shape[0]

        # set explained variance
        self.explained_variance = old_div(d.sum(), vartot)

        # store the eigenvalues
        self.d = d
        # store the eigenvectors
        self.v = v
        # store the total variance
        self.total_variance = vartot

    def get_projmatrix(self, transposed=1):
        """Return the projection matrix."""
        self._if_training_stop_training()
        if transposed:
            return self.v
        return self.v.T

    def get_recmatrix(self, transposed=1):
        """Return the back-projection matrix (i.e. the reconstruction matrix).
        """
        self._if_training_stop_training()
        if transposed:
            return self.v.T
        return self.v

    def _execute(self, x, n=None):
        """Project the input on the first 'n' principal components.
        If 'n' is not set, use all available components."""
        if n is not None:
            return mult(x-self.avg, self.v[:, :n])
        return mult(x-self.avg, self.v)

    def _inverse(self, y, n=None):
        """Project 'y' to the input space using the first 'n' components.
        If 'n' is not set, use all available components."""
        if n is None:
            n = y.shape[1]
        if n > self.output_dim:
            error_str = ("y has dimension %d,"
                         " should be at most %d" % (n, self.output_dim))
            raise mdp.NodeException(error_str)

        v = self.get_recmatrix()
        if n is not None:
            return mult(y, v[:n, :]) + self.avg
        return mult(y, v) + self.avg


class WhiteningNode(PCANode):
    """*Whiten* the input data by filtering it through the most
    significatives of its principal components. All output
    signals have zero mean, unit variance and are decorrelated.

    **Internal variables of interest**

      ``self.avg``
          Mean of the input data (available after training).

      ``self.v``
          Transpose of the projection matrix (available after training).

      ``self.d``
          Variance corresponding to the PCA components (eigenvalues of the
          covariance matrix).

      ``self.explained_variance``
          When output_dim has been specified as a fraction of the total
          variance, this is the fraction of the total variance that is actually
          explained.
    """

    def _stop_training(self, debug=False):
        super(WhiteningNode, self)._stop_training(debug)

        ##### whiten the filters
        # self.v is now the _whitening_ matrix
        self.v = old_div(self.v, numx.sqrt(self.d))

    def get_eigenvectors(self):
        """Return the eigenvectors of the covariance matrix."""
        self._if_training_stop_training()
        return numx.sqrt(self.d)*self.v

    def get_recmatrix(self, transposed=1):
        """Return the back-projection matrix (i.e. the reconstruction matrix).
        """
        self._if_training_stop_training()
        v_inverse = self.v*self.d
        if transposed:
            return v_inverse.T
        return v_inverse



import mdp

def test_Node_deepcopy_lambda():
    """Copying a node with a lambda member function
    should not throw an Exception"""
    generic_node = mdp.Node()
    generic_node.lambda_function = lambda: 1
    generic_node.copy()

def test_Flow_deepcopy_lambda():
    """Copying a Flow with a lambda member function
    should not throw an Exception"""
    generic_node = mdp.Node()
    generic_node.lambda_function = lambda: 1
    generic_flow = mdp.Flow([generic_node])
    generic_flow.copy()



from builtins import str
__docformat__ = "restructuredtext en"

from mdp import numx, numx_linalg, utils, Node, NodeException, TrainingException
from mdp.utils import mult

# ??? For the future: add an optional second phase to compute
# residuals, significance of the slope.

class LinearRegressionNode(Node):
    """Compute least-square, multivariate linear regression on the input
    data, i.e., learn coefficients ``b_j`` so that::

      y_i = b_0 + b_1 x_1 + ... b_N x_N ,

    for ``i = 1 ... M``, minimizes the square error given the training ``x``'s
    and ``y``'s.

    This is a supervised learning node, and requires input data ``x`` and
    target data ``y`` to be supplied during training (see ``train``
    docstring).

    **Internal variables of interest**

      ``self.beta``
          The coefficients of the linear regression
    """

    def __init__(self, with_bias=True, use_pinv=False,
                 input_dim=None, output_dim=None, dtype=None):
        """
        :Arguments:

          with_bias
            If true, the linear model includes a constant term

            - True:  y_i = b_0 + b_1 x_1 + ... b_N x_N
            - False: y_i =       b_1 x_1 + ... b_N x_N

            If present, the constant term is stored in the first
            column of ``self.beta``.

          use_pinv
            If true, uses the pseudo-inverse function to compute
            the linear regression coefficients, which is more robust
            in some cases
        """
        super(LinearRegressionNode, self).__init__(input_dim, output_dim, dtype)

        self.with_bias = with_bias
        self.use_pinv = use_pinv

        # for the linear regression estimator we need two terms
        # the first one is X^T X
        self._xTx = None
        # the second one is X^T Y
        self._xTy = None

        # keep track of how many data points have been sent
        self._tlen = 0

        # final regression coefficients
        # if with_bias=True, beta includes the bias term in the first column
        self.beta = None

    @staticmethod
    def is_invertible():
        return False

    def _check_train_args(self, x, y):
        # set output_dim if necessary
        if self._output_dim is None:
            self._set_output_dim(y.shape[1])
        # check output dimensionality
        self._check_output(y)
        if y.shape[0] != x.shape[0]:
            msg = ("The number of output points should be equal to the "
                   "number of datapoints (%d != %d)" % (y.shape[0], x.shape[0]))
            raise TrainingException(msg)

    def _train(self, x, y):
        """
        **Additional input arguments**

        y
          array of size (x.shape[0], output_dim) that contains the observed
          output to the input x's.
        """
        # initialize internal vars if necessary
        if self._xTx is None:
            if self.with_bias:
                x_size = self._input_dim + 1
            else:
                x_size = self._input_dim
            self._xTx = numx.zeros((x_size, x_size), self._dtype)
            self._xTy = numx.zeros((x_size, self._output_dim), self._dtype)

        if self.with_bias:
            x = self._add_constant(x)

        # update internal variables
        self._xTx += mult(x.T, x)
        self._xTy += mult(x.T, y)
        self._tlen += x.shape[0]

    def _stop_training(self):
        try:
            if self.use_pinv:
                invfun = utils.pinv
            else:
                invfun = utils.inv
            inv_xTx = invfun(self._xTx)
        except numx_linalg.LinAlgError as exception:
            errstr = (str(exception) +
                      "\n Input data may be redundant (i.e., some of the " +
                      "variables may be linearly dependent).")
            raise NodeException(errstr)

        self.beta = mult(inv_xTx, self._xTy)

        # remove junk
        del self._xTx
        del self._xTy

    def _execute(self, x):
        if self.with_bias:
            x = self._add_constant(x)
        return mult(x, self.beta)

    def _add_constant(self, x):
        """Add a constant term to the vector 'x'.
        x -> [1 x]
        """
        return numx.concatenate((numx.ones((x.shape[0], 1),
                                           dtype=self.dtype), x), axis=1)



from __future__ import division
from builtins import object
from past.utils import old_div
import mdp
from .routines import refcast
numx = mdp.numx
numx_linalg = mdp.numx_linalg

# 10 times machine eps
epsilon = 10*numx.finfo(numx.double).eps

class QuadraticFormException(mdp.MDPException):
    pass

class QuadraticForm(object):
    """
    Define an inhomogeneous quadratic form as 1/2 x'Hx + f'x + c .
    This class implements the quadratic form analysis methods
    presented in:

    Berkes, P. and Wiskott, L. (2006). On the analysis and interpretation
    of inhomogeneous quadratic forms as receptive fields. Neural
    Computation, 18(8): 1868-1895.
    """

    def __init__(self, H, f=None, c=None, dtype='d'):
        """
        The quadratic form is defined as 1/2 x'Hx + f'x + c .
        'dtype' specifies the numerical type of the internal structures.
        """
        local_eps = 10*numx.finfo(numx.dtype(dtype)).eps
        # check that H is almost symmetric
        if not numx.allclose(H, H.T, rtol=100*local_eps, atol=local_eps):
            raise QuadraticFormException('H does not seem to be symmetric')

        self.H = refcast(H, dtype)
        if f is None:
            f = numx.zeros((H.shape[0],), dtype=dtype)
        if c is None:
            c = 0
        self.f = refcast(f, dtype)
        self.c = c
        self.dtype = dtype

    def apply(self, x):
        """Apply the quadratic form to the input vectors.
        Return 1/2 x'Hx + f'x + c ."""
        x = numx.atleast_2d(x)
        return (0.5*(mdp.utils.mult(x, self.H.T)*x).sum(axis=1) +
               mdp.utils.mult(x, self.f) + self.c)

    def _eig_sort(self, x):
        E, W = numx_linalg.eig(x)
        E, W = E.real, W.real
        idx = E.argsort()
        E = E.take(idx)
        W = W.take(idx, axis=1)
        return E, W

    def get_extrema(self, norm, tol = 1.E-4):
        """
        Find the input vectors xmax and xmin with norm 'nrm' that maximize
        or minimize the quadratic form.

        tol: norm error tolerance
        """
        H, f, c = self.H, self.f, self.c
        if max(abs(f)) < numx.finfo(self.dtype).eps:
            E, W = self._eig_sort(H)
            xmax = W[:, -1]*norm
            xmin = W[:, 0]*norm
        else:
            H_definite_positive, H_definite_negative = False, False
            E, W = self._eig_sort(H)
            if E[0] >= 0:
                # H is positive definite
                H_definite_positive = True
            elif E[-1] <= 0:
                # H is negative definite
                H_definite_negative = True

            x0 = mdp.numx_linalg.solve(-H, f)
            if H_definite_positive and mdp.utils.norm2(x0) <= norm:
                xmin = x0
                # x0 is a minimum
            else:
                xmin = self._maximize(norm, tol, factor=-1)
            if H_definite_negative and mdp.utils.norm2(x0) <= norm :
                xmax = x0
                # x0 is a maximum
            else:
                xmax = self._maximize(norm, tol, factor=None)

        self.xmax, self.xmin = xmax, xmin
        return xmax, xmin

    def _maximize(self, norm, tol = 1.E-4, x0 = None, factor = None):
        H, f = self.H, self.f
        if factor is not None:
            H = factor*H
            f = factor*f
        if x0 is not None:
            x0 = mdp.utils.refcast(x0, self.dtype)
            f = mdp.utils.mult(H, x0)+ f
            # c = 0.5*x0'*H*x0 + f'*x0 + c -> do we need it?
        mu, V = self._eig_sort(H)
        alpha = mdp.utils.mult(V.T, f).reshape((H.shape[0],))
        # v_i = alpha_i * v_i (alpha is a raw_vector)
        V = V*alpha
        # left bound for lambda
        ll = mu[-1] # eigenvalue's maximum
        # right bound for lambda
        lr = old_div(mdp.utils.norm2(f),norm) + ll
        # search by bisection until norm(x)**2 = norm**2
        norm_2 = norm**2
        norm_x_2 = 0
        while abs(norm_x_2-norm_2) > tol and old_div((lr-ll),lr) > epsilon:
            # bisection of the lambda-interval
            lambd = 0.5*(lr-ll)+ll
            # eigenvalues of (lambda*Id - H)^-1
            beta = (lambd-mu)**(-1)
            # solution to the second lagragian equation
            norm_x_2 = (alpha**2*beta**2).sum()
            #%[ll,lr]
            if norm_x_2 > norm_2:
                ll = lambd
            else:
                lr = lambd
        x = (V*beta).sum(axis=1)
        if x0:
            x = x + x0
        return x

    def get_invariances(self, xstar):
        """Compute invariances of the quadratic form at extremum 'xstar'.
        Outputs:

         w  -- w[:,i] is the direction of the i-th invariance
         nu -- nu[i] second derivative on the sphere in the direction w[:,i]
        """

        # find a basis for the tangential plane of the sphere in x+
        # e(1) ... e(N) is the canonical basis for R^N
        r = mdp.utils.norm2(xstar)
        P = numx.eye(xstar.shape[0], dtype=xstar.dtype)
        P[:, 0] = xstar
        Q, R = numx_linalg.qr(P)
        # the orthogonal subspace
        B = Q[:, 1:]
        # restrict the matrix H to the tangential plane
        Ht = mdp.utils.mult(B.T, mdp.utils.mult(self.H, B))
        # compute the invariances
        nu, w = self._eig_sort(Ht)
        nu -= old_div(((mdp.utils.mult(self.H, xstar)*xstar).sum()
               +(self.f*xstar).sum()),(r*r))
        idx = abs(nu).argsort()
        nu = nu[idx]
        w = w[:, idx]
        w = mdp.utils.mult(B, w)
        return w, nu



"""
Classes for tracing BiNode behavior in flows.
"""
from __future__ import print_function
from builtins import range

import mdp
from bimdp.nodes import IdentityBiNode


class JumpBiNode(IdentityBiNode):
    """BiNode which can perform all kinds of jumps.

    This is useful for testing or flow control. It can also be used
    together with BiNodes as simple jump targets.
    """

    def __init__(self, train_results=None, stop_train_results=None,
                 execute_results=None, *args, **kwargs):
        """Initialize this BiNode.

        Note that this node has an internal variable self.loop_counter which is
        used by execute, message and stop_message (and incremented by each).

        train_results -- List of lists of results for the training phases.
            First index for training phase, second for loop counter.
        stop_train_results -- List of results for the training phases.
        execute_results -- Single result tuple starting at msg or list of
            results, which are used according to the loop counter. The list
            entries can also be None (then x is simply forwarded).
        stop_message_results -- Like execute_results.
        """
        self.loop_counter = 0 # counter for execution phase
        self._train_results = train_results
        self._stop_train_results = stop_train_results
        self._execute_results = execute_results
        super(JumpBiNode, self).__init__(*args, **kwargs)

    def is_trainable(self):
        if self._train_results:
            return True
        else:
            return False

    @staticmethod
    def is_invertible():
        return False

    def _get_train_seq(self):
        """Return a train_seq which returns the predefined values."""
        # wrapper function for _train, using local scopes
        def get_train_function(i_phase):
            def train_function(x):
                self.loop_counter += 1
                if self.loop_counter-1 >= len(self._train_results[i_phase]):
                    return None
                return self._train_results[i_phase][self.loop_counter-1]
            return train_function
        # wrapper function for _stop_training
        def get_stop_training(i_phase):
            def stop_training():
                return self._stop_train_results[i_phase]
            return stop_training
        # now wrap the training sequence
        train_seq = []
        if not self._train_results:
            return train_seq
        for i_phase in range(len(self._train_results)):
            train_seq.append((get_train_function(i_phase),
                              get_stop_training(i_phase)))
        return train_seq

    def _execute(self, x):
        """Return the predefined values for the current loop count value."""
        self.loop_counter += 1
        if not self._execute_results:
            return x
        if self.loop_counter-1 >= len(self._execute_results):
            return x
        result = self._execute_results[self.loop_counter-1]
        if result is None:
            return x
        else:
            return result

    def _bi_reset(self):
        """Reset the loop counter."""
        self.loop_counter = 0

    def is_bi_learning(self):
        return False


class TraceJumpBiNode(JumpBiNode):
    """Node for testing, that logs when and how it is called."""

    def __init__(self, tracelog, log_data=False, verbose=False,
                 *args, **kwargs):
        """Initialize the node.

        tracelog -- list to which to append the log entries
        log_data -- if true the data will be logged as well
        """
        self._tracelog = tracelog
        self._log_data = log_data
        self._verbose = verbose
        super(TraceJumpBiNode, self).__init__(*args, **kwargs)

    def train(self, x, msg=None):
        if self._log_data:
            self._tracelog.append((self._node_id, "train", x, msg))
        else:
            self._tracelog.append((self._node_id, "train"))
        if self._verbose:
            print(self._tracelog[-1])
        return super(TraceJumpBiNode, self).train(x, msg)

    def execute(self, x, msg=None):
        if self._log_data:
            self._tracelog.append((self._node_id, "execute", x, msg))
        else:
            self._tracelog.append((self._node_id, "execute"))
        if self._verbose:
            print(self._tracelog[-1])
        return super(TraceJumpBiNode, self).execute(x, msg)

    def stop_training(self, msg=None):
        self._tracelog.append((self._node_id, "stop_training"))
        if self._verbose:
            print(self._tracelog[-1])
        return super(TraceJumpBiNode, self).stop_training(msg)

    def _bi_reset(self):
        self._tracelog.append((self._node_id, "bi_reset"))
        if self._verbose:
            print(self._tracelog[-1])
        return super(TraceJumpBiNode, self)._bi_reset()


class ParallelTraceJumpBiNode(TraceJumpBiNode):

    def _fork(self):
        return self.copy()

    def _join(self):
        pass


class IdNode(mdp.Node):
    """Non-bi identity node for testing."""

    @staticmethod
    def is_trainable():
        return False



from builtins import map
from builtins import object
# inspired by some code by Nathan Denny (1999)
# see http://www.ece.arizona.edu/~denny/python_nest/graph_lib_1.0.1.html

try:
    # use reduce against BDFL's will even on python > 2.6
    from functools import reduce
except ImportError:
    pass

class GraphException(Exception):
    """Base class for exception in the graph package."""
    pass

class GraphTopologicalException(GraphException):
    """Exception thrown during a topological sort if the graph is cyclical."""
    pass



def is_sequence(x):
    return isinstance(x, (list, tuple))

def recursive_map(func, seq):
    """Apply a function recursively on a sequence and all subsequences."""
    def _func(x):
        if is_sequence(x):
            return recursive_map(func, x)
        else:
            return func(x)
    return list(map(_func, seq))

def recursive_reduce(func, seq, *argv):
    """Apply reduce(func, seq) recursively to a sequence and all its
    subsequences."""
    def _func(x, y):
        if is_sequence(y):
            return func(x, recursive_reduce(func, y))
        else:
            return func(x, y)
    return reduce(_func, seq, *argv)



class GraphNode(object):
    """Represent a graph node and all information attached to it."""

    def __init__(self, data=None):
        self.data = data
        # edges in
        self.ein = []
        # edges out
        self.eout = []

    def add_edge_in(self, edge):
        self.ein.append(edge)

    def add_edge_out(self, edge):
        self.eout.append(edge)

    def remove_edge_in(self, edge):
        self.ein.remove(edge)

    def remove_edge_out(self, edge):
        self.eout.remove(edge)

    def get_edges_in(self, from_ = None):
        """Return a copy of the list of the entering edges. If from_
        is specified, return only the nodes coming from that node."""
        inedges = self.ein[:]
        if from_:
            inedges = [edge for edge in inedges if edge.head == from_]
        return inedges

    def get_edges_out(self, to_ = None):
        """Return a copy of the list of the outgoing edges. If to_
        is specified, return only the nodes going to that node."""
        outedges = self.eout[:]
        if to_:
            outedges = [edge for edge in outedges if edge.tail == to_]
        return outedges

    def get_edges(self, neighbor = None):
        """Return a copy of all edges. If neighbor is specified, return
        only the edges connected to that node."""
        return ( self.get_edges_in(from_=neighbor) +
                 self.get_edges_out(to_=neighbor) )

    def in_degree(self):
        """Return the number of entering edges."""
        return len(self.ein)

    def out_degree(self):
        """Return the number of outgoing edges."""
        return len(self.eout)

    def degree(self):
        """Return the number of edges."""
        return self.in_degree()+self.out_degree()

    def in_neighbors(self):
        """Return the neighbors down in-edges (i.e. the parents nodes)."""
        return [x.get_head() for x in self.ein]

    def out_neighbors(self):
        """Return the neighbors down in-edges (i.e. the parents nodes)."""
        return [x.get_tail() for x in self.eout]

    def neighbors(self):
        return self.in_neighbors() + self.out_neighbors()

class GraphEdge(object):
    """Represent a graph edge and all information attached to it."""

    def __init__(self, head, tail, data=None):
        # head node
        self.head = head
        # neighbors out
        self.tail = tail
        # arbitrary data slot
        self.data = data

    def get_ends(self):
        """Return the tuple (head_id, tail_id)."""
        return (self.head, self.tail)

    def get_tail(self):
        return self.tail

    def get_head(self):
        return self.head

class Graph(object):
    """Represent a directed graph."""

    def __init__(self):
        # list of nodes
        self.nodes = []
        # list of edges
        self.edges = []

    # node functions
    def add_node(self, data=None):
        node = GraphNode(data=data)
        self.nodes.append(node)
        return node

    def remove_node(self, node):
        # the node is not in this graph
        if node not in self.nodes:
            errstr = 'This node is not part of the graph (%s)' % node
            raise GraphException(errstr)

        # remove all edges containing this node
        for edge in node.get_edges():
            self.remove_edge(edge)
        # remove the node
        self.nodes.remove(node)

    # edge functions
    def add_edge(self, head, tail, data=None):
        """Add an edge going from head to tail.
        head : head node
        tail : tail node
        """
        # create edge
        edge = GraphEdge(head, tail, data=data)
        # add edge to head and tail node
        head.add_edge_out(edge)
        tail.add_edge_in(edge)
        # add to the edges dictionary
        self.edges.append(edge)
        return edge

    def remove_edge(self, edge):
        head, tail = edge.get_ends()
        # remove from head
        head.remove_edge_out(edge)
        # remove from tail
        tail.remove_edge_in(edge)
        # remove the edge
        self.edges.remove(edge)

    ### populate functions

    def add_nodes(self, data):
        """Add many nodes at once.

        data -- number of nodes to add or sequence of data values, one for
                each new node"""
        if not is_sequence(data):
            data = [None]*data
        return list(map(self.add_node, data))

    def add_tree(self, tree):
        """Add a tree to the graph.

        The tree is specified with a nested list of tuple, in a LISP-like
        notation. The values specified in the list become the values of
        the single nodes.

        Return an equivalent nested list with the nodes instead of the values.

        Example:
        >>> a=b=c=d=e=None
        >>> g.add_tree( (a, b, (c, d ,e)) )
        corresponds to this tree structure, with all node values set to None:

                a
               / \
              b   c
                 / \
                d   e
        """

        def _add_edge(root, son):
            self.add_edge(root, son)
            return root

        nodes = recursive_map(self.add_node, tree)
        recursive_reduce(_add_edge, nodes)
        return nodes

    def add_full_connectivity(self, from_nodes, to_nodes):
        """Add full connectivity from a group of nodes to another one.
        Return a list of lists of edges, one for each node in 'from_nodes'.

        Example: create a two-layer graph with full connectivity.
        >>> g = Graph()
        >>> layer1 = g.add_nodes(10)
        >>> layer2 = g.add_nodes(5)
        >>> g.add_full_connectivity(layer1, layer2)
        """
        edges = []
        for from_ in from_nodes:
            edges.append([self.add_edge(from_, x) for x in to_nodes])
        return edges

    ###### graph algorithms

    def topological_sort(self):
        """Perform a topological sort of the nodes. If the graph has a cycle,
        throw a GraphTopologicalException with the list of successfully
        ordered nodes."""
        # topologically sorted list of the nodes (result)
        topological_list = []
        # queue (fifo list) of the nodes with in_degree 0
        topological_queue = []
        # {node: in_degree} for the remaining nodes (those with in_degree>0)
        remaining_indegree = {}

        # init queues and lists
        for node in self.nodes:
            indegree = node.in_degree()
            if indegree == 0:
                topological_queue.append(node)
            else:
                remaining_indegree[node] = indegree

        # remove nodes with in_degree 0 and decrease the in_degree of their sons
        while len(topological_queue):
            # remove the first node with degree 0
            node = topological_queue.pop(0)
            topological_list.append(node)
            # decrease the in_degree of the sons
            for son in node.out_neighbors():
                remaining_indegree[son] -= 1
                if remaining_indegree[son] == 0:
                    topological_queue.append(son)

        # if not all nodes were covered, the graph must have a cycle
        # raise a GraphTopographicalException
        if len(topological_list)!=len(self.nodes):
            raise GraphTopologicalException(topological_list)

        return topological_list

    ### Depth-First sort

    def _dfs(self, neighbors_fct, root, visit_fct=None):
        # core depth-first sort function
        # changing the neighbors function to return the sons of a node,
        # its parents, or both one gets normal dfs, reverse dfs, or
        # dfs on the equivalent undirected graph, respectively

        # result list containing the nodes in Depth-First order
        dfs_list = []
        # keep track of all already visited nodes
        visited_nodes = { root: None }

        # stack (lifo) list
        dfs_stack = []
        dfs_stack.append(root)

        while len(dfs_stack):
            # consider the next node on the stack
            node = dfs_stack.pop()
            dfs_list.append(node)
            # visit the node
            if visit_fct != None:
                visit_fct(node)
            # add all sons to the stack (if not already visited)
            for son in neighbors_fct(node):
                if son not in visited_nodes:
                    visited_nodes[son] = None
                    dfs_stack.append(son)

        return dfs_list

    def dfs(self, root, visit_fct=None):
        """Return a list of nodes in some Depth First order starting from
        a root node. If defined, visit_fct is applied on each visited node.

        The returned list does not have to contain all nodes in the
        graph, but only the ones reachable from the root.
        """

        neighbors_fct = lambda node: node.out_neighbors()
        return self._dfs(neighbors_fct, root, visit_fct=visit_fct)

    def undirected_dfs(self, root, visit_fct=None):
        """Perform Depth First sort.

        This function is identical to dfs, but the sort is performed on
        the equivalent undirected version of the graph."""
        neighbors_fct = lambda node: node.neighbors()
        return self._dfs(neighbors_fct, root, visit_fct=visit_fct)

    ### Connected components

    def connected_components(self):
        """Return a list of lists containing the nodes of all connected
        components of the graph."""

        visited = {}
        def visit_fct(node, visited=visited):
            visited[node] = None

        components = []
        nodes = self.nodes
        for node in nodes:
            if node in visited:
                continue
            components.append(self.undirected_dfs(node, visit_fct))
        return components

    def is_weakly_connected(self):
        """Return True if the graph is weakly connected."""
        return len(self.undirected_dfs(self.nodes[0]))==len(self.nodes)

    ### Breadth-First Sort
    # BFS and DFS could be generalized to one function. I leave them
    # distinct for clarity.

    def _bfs(self, neighbors_fct, root, visit_fct=None):
        # core breadth-first sort function
        # changing the neighbors function to return the sons of a node,
        # its parents, or both one gets normal bfs, reverse bfs, or
        # bfs on the equivalent undirected graph, respectively

        # result list containing the nodes in Breadth-First order
        bfs_list = []
        # keep track of all already visited nodes
        visited_nodes = { root: None }

        # queue (fifo) list
        bfs_queue = []
        bfs_queue.append(root)

        while len(bfs_queue):
            # consider the next node in the queue
            node = bfs_queue.pop(0)
            bfs_list.append(node)
            # visit the node
            if visit_fct != None:
                visit_fct(node)
            # add all sons to the queue (if not already visited)
            for son in neighbors_fct(node):
                if son not in visited_nodes:
                    visited_nodes[son] = None
                    bfs_queue.append(son)

        return bfs_list

    def bfs(self, root, visit_fct=None):
        """Return a list of nodes in some Breadth First order starting from
        a root node. If defined, visit_fct is applied on each visited node.

        Note the returned list does not have to contain all nodes in the
        graph, but only the ones reachable from the root."""

        neighbors_fct = lambda node: node.out_neighbors()
        return self._bfs(neighbors_fct, root, visit_fct=visit_fct)

    def undirected_bfs(self, root, visit_fct=None):
        """Perform Breadth First sort.

        This function is identical to bfs, but the sort is performed on
        the equivalent undirected version of the graph."""

        neighbors_fct = lambda node: node.neighbors()
        return self._bfs(neighbors_fct, root, visit_fct=visit_fct)



from ._tools import *

def test_CutoffNode():
    node = mdp.nodes.CutoffNode(-1.5, 1.2)
    x = numx.array([[0.1, 0, -2, 3, 1.2, -1.5, -3.33]])
    y_ref = numx.array([[0.1, 0, -1.5, 1.2, 1.2, -1.5, -1.5]])
    y = node.execute(x)
    assert numx.all(y==y_ref)



from builtins import str
import mdp
from ._tools import *
uniform = mdp.numx_rand.random

def pytest_generate_tests(metafunc):
    _fastica_test_factory(metafunc)

def _fastica_test_factory(metafunc):
    # generate FastICANode testcases
    fica_parm = {'approach': ['symm', 'defl'],
                 'g': ['pow3', 'tanh', 'gaus', 'skew'],
                 'fine_g': [None, 'pow3', 'tanh', 'gaus', 'skew'],
                 'sample_size': [1, 0.99999],
                 'mu': [1, 0.999999],
                 'coarse_limit': [None, 0.01],
                 }

    for parms in mdp.utils.orthogonal_permutations(fica_parm):
        # skew nonlinearity works only with skewed input data
        if parms['g'] != 'skew' and parms['fine_g'] == 'skew':
            continue
        if parms['g'] == 'skew' and parms['fine_g'] != 'skew':
            continue
        # skip testing coarse_limit if g's are the same --
        # it should not be in effect
        if parms['coarse_limit'] is not None \
          and parms['g'] == parms['fine_g']:
          continue
        funcargs = dict(parms=parms)
        theid = fastICA_id(parms)
        metafunc.addcall(funcargs, id=theid)

def fastICA_id(parms):
    app =     'AP:'+parms['approach']
    nl =      'NL:'+parms['g']
    fine_nl = 'FT:'+str(parms['fine_g'])
    coarse_l = 'CL:'+str(parms['coarse_limit'])
    if parms['sample_size'] == 1:
        compact = 'SA:01 '
    else:
        compact = 'SA:<1 '
    if parms['mu'] == 1:
        compact += 'S:01'
    else:
        compact += 'S:<1'
    desc = ' '.join([app, nl, fine_nl, compact, coarse_l])
    return desc


def test_FastICA(parms):
    if parms['g'] == 'skew':
        rand_func = mdp.numx_rand.exponential
    else:
        rand_func = uniform

    # try three times just to clear failures due to randomness
    for exc in (Exception, Exception, ()):
        try:
            ica = mdp.nodes.FastICANode(limit=10**(-decimal),**parms)
            ica2 = ica.copy()
            verify_ICANode(ica, rand_func=rand_func, vars=2)
            verify_ICANodeMatrices(ica2, rand_func=rand_func, vars=2)
        except exc:
            pass



import tempfile
import os
import py.test

def test_tmpdir_exists():
    assert os.path.exists(py.test.mdp_tempdirname)

def test_tmpdir_writable1():
    with open(os.path.join(py.test.mdp_tempdirname, 'empty'), 'w'):
        pass

def test_tmpdir_writable2():
    with tempfile.NamedTemporaryFile(prefix='MDP_', suffix='.testfile',
                                     dir=py.test.mdp_tempdirname):
        pass



from builtins import str
from builtins import range
from ._tools import *

def testPCANode():
    line_x = numx.zeros((1000,2),"d")
    line_y = numx.zeros((1000,2),"d")
    line_x[:,0] = numx.linspace(-1,1,num=1000,endpoint=1)
    line_y[:,1] = numx.linspace(-0.2,0.2,num=1000,endpoint=1)
    mat = numx.concatenate((line_x,line_y))
    des_var = std(mat,axis=0)
    utils.rotate(mat,uniform()*2*numx.pi)
    mat += uniform(2)
    pca = mdp.nodes.PCANode()
    pca.train(mat)
    act_mat = pca.execute(mat)
    assert_array_almost_equal(mean(act_mat,axis=0),\
                              [0,0],decimal)
    assert_array_almost_equal(std(act_mat,axis=0),\
                              des_var,decimal)
    # test that the total_variance attribute makes sense
    est_tot_var = ((des_var**2)*2000/1999.).sum()
    assert_almost_equal(est_tot_var, pca.total_variance, decimal)
    assert_almost_equal(1, pca.explained_variance, decimal)
    # test a bug in v.1.1.1, should not crash
    pca.inverse(act_mat[:,:1])

##     # test segmentation fault with symeig, see
##     # http://projects.scipy.org/scipy/numpy/ticket/551
##     def testPCANode_pickled():
##         for i in xrange(2,100):
##             mat, mix, inp = get_random_mix(mat_dim=(200, i))

##             pca = mdp.nodes.PCANode()
##             pca.train(mat)
##             s = cPickle.dumps(pca)
##             pca = cPickle.loads(s)
##             act_mat = pca.execute(mat)

def testPCANode_total_variance():
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    des_var = ((std(mat, axis=0)**2)*1000/999.).sum()
    pca = mdp.nodes.PCANode(output_dim=2)
    pca.train(mat)
    pca.execute(mat)
    assert_almost_equal(des_var, pca.total_variance, decimal)

def testPCANode_desired_variance():
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    # first make them white
    pca = mdp.nodes.WhiteningNode()
    pca.train(mat)
    mat = pca.execute(mat)
    # set the variances
    mat *= [0.6,0.3,0.1]
    #mat -= mat.mean(axis=0)
    pca = mdp.nodes.PCANode(output_dim=0.8)
    pca.train(mat)
    out = pca.execute(mat)
    # check that we got exactly two output_dim:
    assert pca.output_dim == 2, '%s'%pca.output_dim
    assert out.shape[1] == 2
    # check that explained variance is > 0.8 and < 1
    assert (pca.explained_variance > 0.8 and pca.explained_variance < 1)


def testPCANode_desired_variance_after_train():
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    # first make them white
    pca = mdp.nodes.WhiteningNode()
    pca.train(mat)
    mat = pca.execute(mat)
    # set the variances
    mat *= [0.6,0.3,0.1]
    #mat -= mat.mean(axis=0)
    pca = mdp.nodes.PCANode()
    pca.train(mat)
    # this was not working before the bug fix
    pca.output_dim = 0.8
    out = pca.execute(mat)
    # check that we got exactly two output_dim:
    assert pca.output_dim == 2
    assert out.shape[1] == 2
    # check that explained variance is > 0.8 and < 1
    assert (pca.explained_variance > 0.8 and pca.explained_variance < 1)

def testPCANode_range_argument():
    node = mdp.nodes.PCANode()
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    y = node.execute(x, n=5)
    assert y.shape[1] == 5

def testPCANode_SVD():
    # it should pass atleast the same test as PCANode
    line_x = numx.zeros((1000,2),"d")
    line_y = numx.zeros((1000,2),"d")
    line_x[:,0] = numx.linspace(-1,1,num=1000,endpoint=1)
    line_y[:,1] = numx.linspace(-0.2,0.2,num=1000,endpoint=1)
    mat = numx.concatenate((line_x,line_y))
    des_var = std(mat,axis=0)
    utils.rotate(mat,uniform()*2*numx.pi)
    mat += uniform(2)
    pca = mdp.nodes.PCANode(svd=True)
    pca.train(mat)
    act_mat = pca.execute(mat)
    assert_array_almost_equal(mean(act_mat,axis=0),\
                              [0,0],decimal)
    assert_array_almost_equal(std(act_mat,axis=0),\
                              des_var,decimal)
    # Now a more difficult test, create singular cov matrices
    # and test that PCANode crashes whereas PCASVDNode doesn't
    mat, mix, inp = get_random_mix(mat_dim=(1000, 100), avg=1E+15)
    # now create a degenerate input
    for i in range(1,100):
        inp[:,i] = inp[:,1].copy()
    # check that standard PCA fails
    pca = mdp.nodes.PCANode()
    pca.train(inp)
    try:
        pca.stop_training()
        raise Exception("PCANode didn't catch singular covariance matrix: degenerate")
    except mdp.NodeException:
        pass
    # now try the SVD version
    pca = mdp.nodes.PCANode(svd=True)
    pca.train(inp)
    pca.stop_training()

    # now check the undetermined case
    mat, mix, inp = get_random_mix(mat_dim=(500, 2))
    inp = inp.T
    pca = mdp.nodes.PCANode()
    pca.train(inp)
    try:
        pca.stop_training()
        raise Exception("PCANode didn't catch singular covariance matrix: undetermined")
    except mdp.NodeException:
        pass
    # now try the SVD version
    pca = mdp.nodes.PCANode(svd=True)
    pca.train(inp)
    pca.stop_training()

    # try using the automatic dimensionality reduction function
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    # first make them decorellated
    pca = mdp.nodes.PCANode()
    pca.train(mat)
    mat = pca.execute(mat)
    mat *= [1E+5,1E-3, 1E-4]
    mat -= mat.mean(axis=0)
    pca = mdp.nodes.PCANode(svd=True,reduce=True, var_rel=1E-2)
    pca.train(mat)
    out = pca.execute(mat)
    # check that we got the only large dimension
    assert_array_almost_equal(mat[:,0].mean(axis=0),out.mean(axis=0),
                              decimal)
    assert_array_almost_equal(mat[:,0].std(axis=0),out.std(axis=0),
                              decimal)

    # second test for automatic dimansionality reduction
    # try using the automatic dimensionality reduction function
    mat, mix, inp = get_random_mix(mat_dim=(1000, 3))
    # first make them decorellated
    pca = mdp.nodes.PCANode()
    pca.train(mat)
    mat = pca.execute(mat)
    mat *= [1E+5,1E-3, 1E-18]
    mat -= mat.mean(axis=0)
    pca = mdp.nodes.PCANode(svd=True,reduce=True,
                            var_abs=1E-8, var_rel=1E-30)
    pca.train(mat)
    out = pca.execute(mat)
    # check that we got the only large dimension
    assert_array_almost_equal(mat[:,:2].mean(axis=0),out.mean(axis=0),
                              decimal)
    assert_array_almost_equal(mat[:,:2].std(axis=0),out.std(axis=0),
                              decimal)


def mock_symeig(x, range=None, overwrite=False):
    if range is None:
        N = x.shape[0]
    else:
        N = range[1]-range[0] + 1
    y = numx.zeros((N,))
    z = numx.zeros((N,N))
    y[0] = -1
    y[-1] = 1
    return y, z

def testPCANode_negative_eigenvalues():
    # should throw an Exception if reduce=False and
    # svd = False and output_dim=None
    pca = mdp.nodes.PCANode(output_dim=None, svd=False, reduce=False)
    pca._symeig = mock_symeig
    pca.train(uniform((10,10)))
    try:
        pca.stop_training()
        assert False, "PCA did not catch negative eigenvalues!"
    except mdp.NodeException as e:
        if "Got negative eigenvalues" in str(e):
            pass
        else:
            raise Exception("PCA did not catch negative eigenvalues!\n"+
                            str(e))
    # if reduce=True, should not throw any Exception,
    # and return output_dim = 1
    pca = mdp.nodes.PCANode(output_dim=None, svd=False, reduce=True)
    pca._symeig = mock_symeig
    pca.train(uniform((10,10)))
    pca.stop_training()
    assert pca.output_dim == 1, 'PCA did not remove non-positive eigenvalues!'
    # if svd=True, should not throw any Exception,
    # and return output_dim = 10
    pca = mdp.nodes.PCANode(output_dim=None, svd=True, reduce=False)
    pca._symeig = mock_symeig
    pca.train(uniform((10,10)))
    pca.stop_training()
    assert pca.output_dim == 10, 'PCA did not remove non-positive eigenvalues!'
    # if output_dim is set, should not throw any Exception,
    # and return the right output_dim
    pca = mdp.nodes.PCANode(output_dim=1, svd=False, reduce=False)
    pca._symeig = mock_symeig
    pca.train(uniform((10,10)))
    pca.stop_training()
    assert pca.output_dim == 1, 'PCA did not remove non-positive eigenvalues!'

def test_PCANode_no_eigenvalues_left():
    mat = numx.zeros((100,4), dtype='d')
    pca = mdp.nodes.PCANode(svd=True, reduce=True)
    pca.train(mat)
    py.test.raises(mdp.NodeException, 'pca.stop_training()')
    



"""
Module to convert a flow into an HTML representation.

This is especially useful for hinet structures.
The code uses the visitor pattern to reach and convert all the nodes in a flow.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object

import tempfile
import os
import webbrowser
import io as StringIO

import mdp

from . import switchboard

# TODO: use <pre>   </pre> for whitespaces?

class NewlineWriteFile(object):
    """Decorator for file-like object.

    Adds a newline character to each line written with write().
    """

    def __init__(self, file_obj):
        """Wrap the given file-like object."""
        self.file_obj = file_obj

    def write(self, str_obj):
        """Write a string to the file object and append a newline character."""
        self.file_obj.write(str(str_obj+ "\n"))

    # forward all other methods
    def __getattr__(self, attr):
        return getattr(self.file_obj, attr)


class HiNetHTMLVisitor(object):
    """Class to convert a hinet flow to HTML.

    This class implements the visitor pattern. It relies on the 'html'
    extension to get custom representations for normal node classes.
    """

    def __init__(self, html_file, show_size=False):
        """Initialize the HMTL converter.

        html_file -- File object into which the representation is written
            (only the write method is used).
        show_size -- Show the approximate memory footprint of all nodes.
        """
        self.show_size = show_size
        self._file = NewlineWriteFile(html_file)

    @mdp.with_extension("html")
    def convert_flow(self, flow):
        """Convert the flow into HTML and write it into the internal file."""
        f = self._file
        self._open_node_env(flow, "flow")
        for node in flow:
            f.write('<tr><td>')
            self._visit_node(node)
            f.write('</td></tr>')
        self._close_node_env(flow, "flow")

    _CSS_FILENAME = "hinet.css"

    @classmethod
    def hinet_css(cls):
        """Return the standard CSS string.

        The CSS should be embedded in the final HTML file.
        """
        css_filename = os.path.join(os.path.split(__file__)[0],
                                    cls._CSS_FILENAME)
        with open(css_filename, 'r') as css_file:
            css = css_file.read()
        return css

    def _visit_node(self, node):
        """Translate a node and return the translation.

        Depending on the type of the node this can be delegated to more
        specific methods.
        """
        if hasattr(node, "flow"):
            self._visit_flownode(node)
        elif isinstance(node, mdp.hinet.CloneLayer):
            self._visit_clonelayer(node)
        elif isinstance(node, mdp.hinet.SameInputLayer):
            self._visit_sameinputlayer(node)
        elif isinstance(node, mdp.hinet.Layer):
            self._visit_layer(node)
        else:
            self._visit_standard_node(node)

    def _visit_flownode(self, flownode):
        f = self._file
        self._open_node_env(flownode, "flownode")
        for node in flownode.flow:
            f.write('<tr><td>')
            self._visit_node(node)
            f.write('</td></tr>')
        self._close_node_env(flownode, "flownode")

    def _visit_layer(self, layer):
        f = self._file
        self._open_node_env(layer, "layer")
        f.write('<tr>')
        for node in layer:
            f.write('<td>')
            self._visit_node(node)
            f.write('</td>')
        f.write('</tr>')
        self._close_node_env(layer)

    def _visit_clonelayer(self, layer):
        f = self._file
        self._open_node_env(layer, "layer")
        f.write('<tr><td class="nodename">')
        f.write(str(layer) + '<br><br>')
        f.write('%d repetitions' % len(layer))
        f.write('</td>')
        f.write('<td>')
        self._visit_node(layer.node)
        f.write('</td></tr>')
        self._close_node_env(layer)

    def _visit_sameinputlayer(self, layer):
        f = self._file
        self._open_node_env(layer, "layer")
        f.write('<tr><td colspan="%d" class="nodename">%s</td></tr>' %
                (len(layer), str(layer)))
        f.write('<tr>')
        for node in layer:
            f.write('<td>')
            self._visit_node(node)
            f.write('</td>')
        f.write('</tr>')
        self._close_node_env(layer)

    def _visit_standard_node(self, node):
        f = self._file
        self._open_node_env(node)
        f.write('<tr><td class="nodename">')
        f.write(str(node))
        f.write('</td></tr>')
        f.write('<tr><td class="nodeparams">')
        f.write(node.html_representation())
        f.write('</td></tr>')
        self._close_node_env(node)

    # helper methods for decoration

    def _open_node_env(self, node, type_id="node"):
        """Open the HTML environment for the node internals.

        node -- The node itself.
        type_id -- The id string as used in the CSS.
        """
        self._file.write('<table class="%s">' % type_id)
        self._write_node_header(node, type_id)

    def _write_node_header(self, node, type_id="node"):
        """Write the header content for the node into the HTML file."""
        f = self._file
        if not (type_id=="flow" or type_id=="flownode"):
            f.write('<tr><td class="dim">in-dim: %s</td></tr>' %
                    str(node.input_dim))
        f.write('<tr><td>')
        f.write('<table class="nodestruct">')

    def _close_node_env(self, node, type_id="node"):
        """Close the HTML environment for the node internals.

        node -- The node itself.
        type_id -- The id string as used in the CSS.
        """
        f = self._file
        f.write('</table>')
        f.write('</td></tr>')
        if not (type_id=="flow" or type_id=="flownode"):
            f.write('<tr><td class="dim">out-dim: %s' % str(node.output_dim))
            if self.show_size:
                f.write('&nbsp;&nbsp;<span class="memorycolor">size: %s</span>'
                        % mdp.utils.get_node_size_str(node))
            f.write('</td></tr>')
        f.write('</table>')


class  HTMLExtensionNode(mdp.ExtensionNode, mdp.Node):
    """Extension node for custom HTML representations of individual nodes.

    This extension works together with the HiNetHTMLVisitor to allow the
    polymorphic generation of representations for node classes.
    """

    extension_name = "html"

    def html_representation(self):
        """Return an HTML representation of the node."""
        html_repr = self._html_representation()
        if type(html_repr) is str:
            return html_repr
        else:
            return " <br>\n".join(html_repr)

    # override this method
    def _html_representation(self):
        """Return either the final HTML code or a list of HTML lines."""
        return ""


@mdp.extension_method("html", switchboard.Rectangular2dSwitchboard,
                      "_html_representation")
def _rect2d_switchoard_html(self):
    lines = ['rec. field size (in channels): %d x %d = %d' %
                (self.field_channels_xy[0], self.field_channels_xy[1],
                 self.field_channels_xy[0] * self.field_channels_xy[1]),
             '# of rec. fields (out channels): %d x %d = %d' %
                (self.out_channels_xy[0], self.out_channels_xy[1],
                 self.output_channels),
             'rec. field distances (in channels): ' +
                str(self.field_spacing_xy),
             'channel width: %d' % self.in_channel_dim]
    if not all(self.unused_channels_xy):
        lines.append('unused channels: ' + str(self.unused_channels_xy))
    return lines

@mdp.extension_method("html", switchboard.DoubleRect2dSwitchboard,
                      "_html_representation")
def _double_rect2d_switchoard_html(self):
    lines = ['rec. field size (in channels): %d x %d = %d' %
                (self.field_channels_xy[0], self.field_channels_xy[1],
                 self.field_channels_xy[0] * self.field_channels_xy[1]),
             '# of long row rec. fields (out channels): ' +
                str(self.long_out_channels_xy),
             'total number of receptive fields: %d' %
                self.output_channels,
             'channel width: %d' % self.in_channel_dim]
    if self.x_unused_channels or self.y_unused_channels:
        lines.append('unused channels: ' + str(self.unused_channels_xy))
    return lines

@mdp.extension_method("html", switchboard.DoubleRhomb2dSwitchboard,
                      "_html_representation")
def _double_rhomb2d_switchoard_html(self):
    lines = ['rec. field size: %d' % self.diag_field_channels,
             '# of rec. fields (out channels): %d x %d = %d' %
                (self.out_channels_xy[0], self.out_channels_xy[1],
                 self.output_channels),
             'channel width: %d' % self.in_channel_dim]
    return lines

@mdp.extension_method("html", mdp.nodes.SFA2Node, "_html_representation")
def _sfa_html(self):
    return 'expansion dim: ' + str(self._expnode.output_dim)

@mdp.extension_method("html", mdp.nodes.NormalNoiseNode,
                      "_html_representation")
def _noise_html(self):
    return ['noise level: ' + str(self.noise_args[1]),
            'noise offset: ' + str(self.noise_args[0])]

@mdp.extension_method("html", mdp.nodes.CutoffNode, "_html_representation")
def _cutoff_html(self):
    return ['lower bound: ' + str(self.lower_bound),
            'upper bound: ' + str(self.upper_bound)]

@mdp.extension_method("html", mdp.nodes.HistogramNode, "_html_representation")
def _hist_html(self):
    return 'history data fraction: ' + str(self.hist_fraction)

@mdp.extension_method("html", mdp.nodes.AdaptiveCutoffNode,
                      "_html_representation")
def _adap_html(self):
    return ['lower cutoff fraction: ' + str(self.lower_cutoff_fraction),
            'upper cutoff fraction: ' + str(self.upper_cutoff_fraction),
            'history data fraction: ' + str(self.hist_fraction)]


class HiNetXHTMLVisitor(HiNetHTMLVisitor):
    """Modified converter to create valid XHTML."""

    def convert_flow(self, flow):
        """Convert the flow into XHTML and write it into the internal file."""
        # first write the normal HTML into a buffer
        orig_file = self._file
        html_file = StringIO.StringIO()
        self._file = NewlineWriteFile(html_file)
        super(HiNetXHTMLVisitor, self).convert_flow(flow)
        self._file = orig_file
        # now convert it to XHTML
        html_code = html_file.getvalue()
        html_code = html_code.replace('<br>', '<br />')
        html_code = html_code.replace('&nbsp;', '&#160;')
        self._file.write(html_code)


## Helper functions ##

def show_flow(flow, filename=None, title="MDP flow display",
              show_size=False, browser_open=True):
    """Write a flow into a HTML file, open it in the browser and
    return the file name.

    flow -- The flow to be shown.
    filename -- Filename for the HTML file to be created. If None
                a temporary file is created.
    title -- Title for the HTML file.
    show_size -- Show the approximate memory footprint of all nodes.
    browser_open -- If True (default value) then the slideshow file is
        automatically opened in a webbrowser.
    """
    if filename is None:
        (fd, filename) = tempfile.mkstemp(suffix=".html", prefix="MDP_")
        html_file = os.fdopen(fd, 'w')
    else:
        html_file = open(filename, 'w')
    html_file.write('<html>\n<head>\n<title>%s</title>\n' % title)
    html_file.write('<style type="text/css" media="screen">')
    html_file.write(mdp.utils.basic_css() + HiNetHTMLVisitor.hinet_css())
    html_file.write('</style>\n</head>\n<body>\n')
    html_file.write('<h3>%s</h3>\n' % title)
    explanation = '(data flows from top to bottom)'
    html_file.write('<par class="explanation">%s</par>\n' % explanation)
    html_file.write('<br><br><br>\n')
    converter = mdp.hinet.HiNetHTMLVisitor(html_file, show_size=show_size)
    converter.convert_flow(flow=flow)
    html_file.write('</body>\n</html>')
    html_file.close()
    if browser_open:
        webbrowser.open(os.path.abspath(filename))
    return filename



from builtins import range
from ._tools import *

import mdp.parallel as parallel
n = numx

def test_tasks():
    """Test parallel training and execution by running the tasks."""
    flow = parallel.ParallelFlow([
                        mdp.nodes.SFANode(output_dim=5),
                        mdp.nodes.PolynomialExpansionNode(degree=3),
                        mdp.nodes.SFANode(output_dim=20)])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    # parallel execution
    iterable = [n.random.random((20,10)) for _ in range(6)]
    flow.execute(iterable, scheduler=scheduler)

def test_non_iterator():
    """Test parallel training and execution with a single array."""
    flow = parallel.ParallelFlow([
                        mdp.nodes.SFANode(output_dim=5),
                        mdp.nodes.PolynomialExpansionNode(degree=3),
                        mdp.nodes.SFANode(output_dim=20)])
    data_iterables = n.random.random((200,10))*n.arange(1,11)
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    # test execution
    x = n.random.random((100,10))
    flow.execute(x)

def test_multiple_schedulers():
    """Test parallel flow training with multiple schedulers."""
    flow = parallel.ParallelFlow([
                        mdp.nodes.SFANode(output_dim=5),
                        mdp.nodes.PolynomialExpansionNode(degree=3),
                        mdp.nodes.SFANode(output_dim=20)])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    schedulers = [parallel.Scheduler(), None, parallel.Scheduler()]
    flow.train(data_iterables, scheduler=schedulers)
    # parallel execution
    iterable = [n.random.random((20,10)) for _ in range(6)]
    flow.execute(iterable, scheduler=parallel.Scheduler())

def test_multiple_schedulers2():
    """Test parallel flow training with multiple schedulers (part 2)."""
    # now the first node is untrainable as well
    flow = parallel.ParallelFlow([
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=5),
                        mdp.nodes.PolynomialExpansionNode(degree=3),
                        mdp.nodes.SFANode(output_dim=20)])
    data_iterables = [None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    schedulers = [None, parallel.Scheduler(), None, parallel.Scheduler()]
    flow.train(data_iterables, scheduler=schedulers)
    # parallel execution
    iterable = [n.random.random((20,10)) for _ in range(6)]
    flow.execute(iterable, scheduler=parallel.Scheduler())

def test_multiphase():
    """Test parallel training and execution for nodes with multiple
    training phases.
    """
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
    flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node, sfa2_node]))
    flow = parallel.ParallelFlow([
                        flownode,
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=5)])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    # test normal execution
    x = n.random.random([100,10])
    flow.execute(x)
    # parallel execution
    iterable = [n.random.random((20,10)) for _ in range(6)]
    flow.execute(iterable, scheduler=scheduler)

def test_firstnode():
    """Test special case in which the first node is untrainable.

    This tests the proper initialization of the internal variables.
    """
    flow = parallel.ParallelFlow([
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=20)])
    data_iterables = [None,
                       n.random.random((6,20,10))]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)

def test_multiphase_checkpoints():
    """Test parallel checkpoint flow."""
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
    flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node, sfa2_node]))
    flow = parallel.ParallelCheckpointFlow([
                        flownode,
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=5)])
    data_iterables = [[n.random.random((30,10)) for _ in range(6)],
                      None,
                      [n.random.random((30,10)) for _ in range(6)]]
    checkpoint = mdp.CheckpointFunction()
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler, checkpoints=checkpoint)

def test_nonparallel1():
    """Test training for mixture of parallel and non-parallel nodes."""
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    # TODO: use a node with no parallel here
    sfa2_node = mdp.nodes.CuBICANode(input_dim=8)
    flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node, sfa2_node]))
    flow = parallel.ParallelFlow([
                        flownode,
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=5)])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    # test execution
    x = n.random.random([100,10])
    flow.execute(x)

def test_nonparallel2():
    """Test training for mixture of parallel and non-parallel nodes."""
    # TODO: use a node with no parallel here
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
    flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node, sfa2_node]))
    flow = parallel.ParallelFlow([
                        flownode,
                        mdp.nodes.PolynomialExpansionNode(degree=2),
                        mdp.nodes.SFANode(output_dim=5)])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      None,
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    # test execution
    x = n.random.random([100,10])
    flow.execute(x)

def test_nonparallel3():
    """Test training for non-parallel nodes."""
    # TODO: use a node with no parallel here
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    # TODO: use a node with no parallel here
    sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
    flow = parallel.ParallelFlow([sfa_node, sfa2_node])
    data_iterables = [[n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)],
                      [n.random.random((30,10))*n.arange(1,11)
                       for _ in range(6)]]
    scheduler = parallel.Scheduler()
    flow.train(data_iterables, scheduler=scheduler)
    while flow.is_parallel_training:
        results = []
        while flow.task_available():
            task = flow.get_task()
            results.append(task())
        flow.use_results(results)
    # test execution
    x = n.random.random([100,10])
    flow.execute(x)
    
def test_train_purge_nodes():
    """Test that FlowTrainCallable correctly purges nodes."""
    sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
    sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
    flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node,
                                            mdp.nodes.IdentityNode(),
                                            sfa2_node]))
    data = n.random.random((30,10))
    mdp.activate_extension("parallel")
    try:
        clbl = mdp.parallel.FlowTrainCallable(flownode)
        flownode = clbl(data)
    finally:
        mdp.deactivate_extension("parallel")
    assert flownode._flow[1].__class__.__name__ == "_DummyNode"
    
def test_execute_fork():
    """Test the forking of a node based on use_execute_fork."""
    
    class _test_ExecuteForkNode(mdp.nodes.IdentityNode):
        
        # Note: The explicit signature is important to preserve the dim
        #    information during the fork.
        def __init__(self, input_dim=None, output_dim=None, dtype=None):
            self.n_forks = 0
            self.n_joins = 0
            super(_test_ExecuteForkNode, self).__init__(input_dim=input_dim,
                                                        output_dim=output_dim,
                                                        dtype=dtype)
    
    class Parallel_test_ExecuteForkNode(parallel.ParallelExtensionNode,
                                        _test_ExecuteForkNode):
       
        def _fork(self):
            self.n_forks += 1
            return self._default_fork()
        
        def _join(self, forked_node):
            self.n_joins += forked_node.n_joins + 1
            
        def use_execute_fork(self):
            return True
    
    try:
        n_chunks = 6
        
        ## Part 1: test execute fork during flow training
        data_iterables = [[n.random.random((30,10)) for _ in range(n_chunks)],
                          None,
                          [n.random.random((30,10)) for _ in range(n_chunks)],
                          None]
        flow = parallel.ParallelFlow([mdp.nodes.PCANode(output_dim=5),
                                      _test_ExecuteForkNode(),
                                      mdp.nodes.SFANode(),
                                      _test_ExecuteForkNode()])
        scheduler = parallel.Scheduler()
        flow.train(data_iterables, scheduler=scheduler)
        for node in flow:
            if isinstance(node, _test_ExecuteForkNode):
                assert node.n_forks == 2 * n_chunks + 2
                assert node.n_joins == 2 * n_chunks
                # reset the counters to prepare the execute test
                node.n_forks = 0
                node.n_joins = 0
                
        ## Part 2: test execute fork during flow execute
        data_iterable = [n.random.random((30,10)) for _ in range(n_chunks)]
        flow.execute(data_iterable, scheduler=scheduler)
        for node in flow:
            if isinstance(node, _test_ExecuteForkNode):
                assert node.n_forks == n_chunks
                assert node.n_joins == n_chunks
    finally:
        # unregister the testing class
        del mdp.get_extensions()["parallel"][_test_ExecuteForkNode]
        scheduler.shutdown()
  




from __future__ import print_function
from builtins import str
__docformat__ = "restructuredtext en"

from .routines import (timediff, refcast, scast, rotate, random_rot,
                      permute, symrand, norm2, cov2,
                      mult_diag, comb, sqrtm, get_dtypes, nongeneral_svd,
                      hermitian, cov_maxima,
                      lrep, rrep, irep, orthogonal_permutations,
                      izip_stretched,
                      weighted_choice, bool_to_sign, sign_to_bool, gabor,
                      invert_exp_funcs2)
try:
    from collections import OrderedDict
except ImportError:
    ## Getting an Ordered Dict for Python < 2.7
    from ._ordered_dict import OrderedDict

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from .temporarydir import TemporaryDirectory

from .introspection import dig_node, get_node_size, get_node_size_str
from .quad_forms import QuadraticForm, QuadraticFormException
from .covariance import (CovarianceMatrix, DelayCovarianceMatrix,
                        MultipleCovarianceMatrices,CrossCovarianceMatrix)
from .progress_bar import progressinfo
from .slideshow import (basic_css, slideshow_css, HTMLSlideShow,
                       image_slideshow_css, ImageHTMLSlideShow,
                       SectionHTMLSlideShow, SectionImageHTMLSlideShow,
                       image_slideshow, show_image_slideshow)

from ._symeig import SymeigException

import mdp as _mdp
# matrix multiplication function
# we use an alias to be able to use the wrapper for the 'gemm' Lapack
# function in the future
mult = _mdp.numx.dot
matmult = mult

if _mdp.numx_description == 'scipy':
    def matmult(a,b, alpha=1.0, beta=0.0, c=None, trans_a=0, trans_b=0):
        """Return alpha*(a*b) + beta*c.
        a,b,c : matrices
        alpha, beta: scalars
        trans_a : 0 (a not transposed), 1 (a transposed),
                  2 (a conjugate transposed)
        trans_b : 0 (b not transposed), 1 (b transposed),
                  2 (b conjugate transposed)
        """
        if c:
            gemm,=_mdp.numx_linalg.get_blas_funcs(('gemm',),(a,b,c))
        else:
            gemm,=_mdp.numx_linalg.get_blas_funcs(('gemm',),(a,b))

        return gemm(alpha, a, b, beta, c, trans_a, trans_b)

# workaround to numpy issues with dtype behavior:
# 'f' is upcasted at least in the following functions
_inv = _mdp.numx_linalg.inv
inv = lambda x: refcast(_inv(x), x.dtype)
_pinv = _mdp.numx_linalg.pinv
pinv = lambda x: refcast(_pinv(x), x.dtype)
_solve = _mdp.numx_linalg.solve
solve = lambda x, y: refcast(_solve(x, y), x.dtype)

def svd(x, compute_uv = True):
    """Wrap the numx SVD routine, so that it returns arrays of the correct
    dtype and a SymeigException in case of failures."""
    tc = x.dtype
    try:
        if compute_uv:
            u, s, v = _mdp.numx_linalg.svd(x)
            return refcast(u, tc), refcast(s, tc), refcast(v, tc)
        else:
            s = _mdp.numx_linalg.svd(x, compute_uv=False)
            return refcast(s, tc)
    except _mdp.numx_linalg.LinAlgError as exc:
        raise SymeigException(str(exc))

__all__ = ['CovarianceMatrix', 'DelayCovarianceMatrix','CrossCovarianceMatrix',
           'MultipleCovarianceMatrices', 'QuadraticForm',
           'QuadraticFormException',
           'comb', 'cov2', 'dig_node', 'get_dtypes', 'get_node_size',
           'hermitian', 'inv', 'mult', 'mult_diag', 'nongeneral_svd',
           'norm2', 'permute', 'pinv', 'progressinfo',
           'random_rot', 'refcast', 'rotate', 'scast', 'solve', 'sqrtm',
           'svd', 'symrand', 'timediff', 'matmult',
           'HTMLSlideShow', 'ImageHTMLSlideShow',
           'basic_css', 'slideshow_css', 'image_slideshow_css',
           'SectionHTMLSlideShow',
           'SectionImageHTMLSlideShow', 'image_slideshow',
           'lrep', 'rrep', 'irep',
           'orthogonal_permutations', 'izip_stretched',
           'weighted_choice', 'bool_to_sign', 'sign_to_bool',
           'OrderedDict', 'TemporaryDirectory', 'gabor', 'fixup_namespace']

def _without_prefix(name, prefix):
    if name.startswith(prefix):
        return name[len(prefix):]
    else:
        return None

import os
FIXUP_DEBUG = os.getenv('MDPNSDEBUG')

def fixup_namespace(mname, names, old_modules, keep_modules=()):
    """Update ``__module__`` attribute and remove ``old_modules`` from namespace

    When classes are imported from implementation modules into the
    package exporting them, the ``__module__`` attribute reflects the
    place of definition. Splitting the code into separate files (and
    thus modules) makes the implementation managable. Nevertheless, we
    do not want the implementation modules to be visible and delete
    their names from the package's namespace. This causes some
    problems: when looking at the exported classes and other objects,
    their ``__module__`` attribute points to something non-importable,
    ``repr`` output and documentation do not show the module from
    which they are supposed to be imported. The documentation
    generators like epydoc and sphinx are also confused. To alleviate
    those problems, the ``__module__`` attributes of all exported
    classes defined in a "private" module and then exported elsewhere
    are changed to the latter.

    For each name in ``names``, if ``<mname>.<name>`` is accessible,
    and if its ``__module__`` attribute is equal to one of the names
    in ``old_modules``, it is changed to ``"<mname>"``. In other
    words, all the ``__module__`` attributes of objects exported from
    module ``<mname>`` are updated, iff they used to point to one of the
    "private" modules in ``old_modules``.

    This operation is performed not only for classes, but actually for
    all objects with the ``__module__`` attribute, following the rules
    stated above. The operation is also performed recursively, not
    only for names in ``names``, but also for methods, inner classes,
    and other attributes. This recursive invocation is necessary
    because all the problems affecting top-level exported classes also
    affect their attributes visible for the user, and especially
    documented functions.

    If ``names`` is ``None``, all public names in module ``<mname>``
    (not starting with ``'_'``) are affected.

    After the ``__module__`` attributes are changed, "private" modules given
    in ``old_modules``, except for the ones in ``keep_modules``, are deleted
    from the namespace of ``<mname>`` module.
    """
    import sys
    module = sys.modules[mname]
    if names is None:
        names = [name for name in dir(module) if not name.startswith('_')]
    if FIXUP_DEBUG:
        print('NAMESPACE FIXUP: %s (%s)' % (module, mname))
    for name in names:
        _fixup_namespace_item(module, mname, name, old_modules, '')

    # take care of removing the module filenames
    for filename in old_modules:
        # skip names in keep modules
        if filename in keep_modules:
            continue
        try:
            delattr(module, filename)
            if FIXUP_DEBUG:
                print('NAMESPACE FIXUP: deleting %s from %s' % (filename, module))
        except AttributeError:
            # if the name is not there, we are in a reload, so do not
            # do anything
            pass

def _fixup_namespace_item(parent, mname, name, old_modules, path):
    try:
        item = getattr(parent, name)
    except AttributeError:
        if name.startswith('__'): # those sometimes fail unexplicably
            return
        else:
            raise
    current_name = getattr(item, '__module__', None)
    if (current_name is not None and
        _without_prefix(current_name, mname + '.') in old_modules):
        if FIXUP_DEBUG:
            print('namespace fixup: {%s => %s}%s.%s' % (
                current_name, mname, path, name))
        try:
            item.__module__ = mname
        except AttributeError:
            try:
                item.__func__.__module__ = mname
            except AttributeError as e:
                if FIXUP_DEBUG:
                    print('namespace fixup failed: ', e)
            # don't recurse into functions anyway
            return
        subitems = [_name for _name in dir(item)
                    if _name.startswith('__') or not _name.startswith('_')]
        for subitem in subitems:
            _fixup_namespace_item(item, mname, subitem, old_modules,
                                  path + '.' + name)

fixup_namespace(__name__, __all__,
                ('routines',
                 'introspection',
                 'quad_forms',
                 'covariance',
                 'progress_bar',
                 'slideshow',
                 '_ordered_dict',
                 'templet',
                 'temporarydir',
                 'os',
                 ))



"""
Extension to get the total derivative / gradient / Jacobian matrix.
"""
from builtins import str
from builtins import range

import mdp
import bimdp
np = mdp.numx

class NotDifferentiableException(mdp.NodeException):
    """Exception if the total derivative does not exist."""
    pass


# Default implementation is needed to satisfy the "method" request.
class GradientExtensionNode(mdp.ExtensionNode, mdp.Node):
    """Base node of the extension to calculate the gradient at a certain point.

    To get the gradient simply put 'method': 'gradient' into the msg dict.

    The grad array is three dimensional, with shape
    (len(x), self.output_dim, self.input_dim).
    The matrix formed by the last two indices is also called the Jacobian
    matrix.

    Nodes which have no well defined total derivative should raise the
    NotDifferentiableException.
    """

    extension_name = "gradient"

    def _gradient(self, x, grad=None):
        """Calculate the contribution to the grad for this node at point x.

        The contribution is then combined with the given gradient, to get
        the gradient for the original x.

        This is a template function, derived classes should override _get_grad.
        """
        if self.is_training():
            raise mdp.TrainingException("The training is not completed yet.")
        if grad is None:
            grad = np.zeros((len(x), self.input_dim, self.input_dim))
            diag_indices = np.arange(self.input_dim)
            grad[:,diag_indices,diag_indices] = 1.0
        new_grad = self._get_grad(x)
        # combine the gradients
        grad = np.asarray([np.dot(new_grad[i], grad[i])
                           for i in range(len(new_grad))])
        # update the x value for the next node
        result = self._execute(x)
        if isinstance(result, tuple):
            x = result[0]
            msg = result[1]
        else:
            x = result
            msg = {}
        msg.update({"grad": grad})
        return x, msg

    def _get_grad(self, x):
        """Return the grad for the given points.

        Override this method.
        """
        err = "Gradient not implemented for class %s." % str(self.__class__)
        raise NotImplementedError(err)

    def _stop_gradient(self, x, grad=None):
        """Helper method to make gradient available for stop_message."""
        result = self._gradient(x, grad)
        # FIXME: Is this really correct? x should be updated!
        #    Could remove this once we have the new stop signature.
        return result[1], 1


## Implementations for specific nodes. ##

# TODO: cache the gradient for linear nodes?
#    If there was a linear base class one could integrate this?

# TODO: add at least a PCA gradient implementation

@mdp.extension_method("gradient", mdp.nodes.IdentityNode, "_get_grad")
def _identity_grad(self, x):
    grad = np.zeros((len(x), self.output_dim, self.input_dim))
    diag_indices = np.arange(self.input_dim)
    grad[:,diag_indices,diag_indices] = 1.0
    return grad

@mdp.extension_method("gradient", mdp.nodes.SFANode, "_get_grad")
def _sfa_grad(self, x):
    # the gradient is constant, but have to give it for each x point
    return np.repeat(self.sf.T[np.newaxis,:,:], len(x), axis=0)

@mdp.extension_method("gradient", mdp.nodes.QuadraticExpansionNode,
                      "_get_grad")
def _quadex_grad(self, x):
    # the exapansion is:
    # [x1, x2, x3, x1x1, x1x2, x1x3, x2x2, x2x3, x3,x3]
    dim = self.input_dim
    grad = np.zeros((len(x), self.output_dim, dim))
    # constant part
    diag_indices = np.arange(dim)
    grad[:,diag_indices,diag_indices] = 1.0
    # quadratic part
    i_start = dim
    for i in range(dim):
        grad[:, i_start:i_start+dim-i, i] = x[:,i:]
        diag_indices = np.arange(dim - i)
        grad[:, diag_indices+i_start, diag_indices+i] += x[:,i,np.newaxis]
        i_start += (dim - i)
    return grad

@mdp.extension_method("gradient", mdp.nodes.SFA2Node, "_get_grad")
def _sfa2_grad(self, x):
    quadex_grad = self._expnode._get_grad(x)
    sfa_grad = _sfa_grad(self, x)
    return np.asarray([np.dot(sfa_grad[i], quadex_grad[i])
                       for i in range(len(sfa_grad))])

## mdp.hinet nodes ##

@mdp.extension_method("gradient", mdp.hinet.Layer, "_get_grad")
def _layer_grad(self, x):
    in_start = 0
    in_stop = 0
    out_start = 0
    out_stop = 0
    grad = None
    for node in self.nodes:
        out_start = out_stop
        out_stop += node.output_dim
        in_start = in_stop
        in_stop += node.input_dim
        if grad is None:
            node_grad = node._get_grad(x[:, in_start:in_stop])
            grad = np.zeros([node_grad.shape[0], self.output_dim,
                             self.input_dim],
                            dtype=node_grad.dtype)
            # note that the gradient is block-diagonal
            grad[:, out_start:out_stop, in_start:in_stop] = node_grad
        else:
            grad[:, out_start:out_stop, in_start:in_stop] = \
                node._get_grad(x[:, in_start:in_stop])
    return grad

# this is an optimized implementation, the original implementation is
# used for reference in the unittest
@mdp.extension_method("gradient", mdp.hinet.Switchboard, "_gradient")
def _switchboard_gradient(self, x, grad=None):
    if grad is None:
        grad = np.zeros((len(x), self.input_dim, self.input_dim))
        diag_indices = np.arange(self.input_dim)
        grad[:,diag_indices,diag_indices] = 1.0
    ## custom implementation for greater speed 
    grad =  grad[:, self.connections]
    # update the x value for the next node
    result = self._execute(x)
    if isinstance(result, tuple):
        x = result[0]
        msg = result[1]
    else:
        x = result
        msg = {}
    msg.update({"grad": grad})
    return x, msg




import mdp
import inspect
X = mdp.numx_rand.random(size=(500,5))

def get_signature(func):
    regargs, varargs, varkwargs, defaults = inspect.getargspec(func)
    return inspect.formatargspec(regargs, varargs, varkwargs, defaults,
                                 formatvalue=lambda value: "")[1:-1]

def test_docstrings():
    # first try on a subclass of Node if
    # the docstring is exported to the public method
    class AncestorNode(mdp.Node):
        def _train(self, x):
            """doc ancestor"""
            self.foo = 42
    anode = AncestorNode()
    assert anode.train.__doc__ == "doc ancestor"
    anode.train(X)
    assert anode.foo == 42
    assert get_signature(anode.train) == 'self, x'


    # now try on a subclass of it
    class ChildNode(AncestorNode):
        def _train(self, x):
            """doc child"""
            self.foo2 = 42
    cnode = ChildNode()
    assert cnode.train.__doc__ == "doc child"
    cnode.train(X)
    assert cnode.foo2 == 42
    assert get_signature(cnode.train) == 'self, x'

def test_signatures_no_doc():
    # first try on a subclass of Node if
    # the signature is exported to the public method
    class AncestorNode(mdp.Node):
        def _train(self, x, foo=None):
            self.foo = 42
    anode = AncestorNode()
    anode.train(X, foo='abc')
    assert anode.foo == 42
    assert get_signature(anode.train) == 'self, x, foo'
    
    # now try on a subclass of it
    class ChildNode(AncestorNode):
        def _train(self, x, foo2=None):
            self.foo2 = 42
    cnode = ChildNode()
    cnode.train(X, foo2='abc')
    assert cnode.foo2 == 42
    assert get_signature(cnode.train) == 'self, x, foo2'

def test_signatures_with_doc_in_both():
    # first try on a subclass of Node if
    # the signature and the docstring are exported to
    # the public method
    class AncestorNode(mdp.Node):
        def _train(self, x, foo=None):
            """doc ancestor"""
            self.foo = 42
    anode = AncestorNode()
    assert anode.train.__doc__ == "doc ancestor"
    anode.train(X, foo='abc')
    assert anode.foo == 42
    assert get_signature(anode.train) == 'self, x, foo'
    
    # now try on a subclass of it
    class ChildNode(AncestorNode):
        def _train(self, x, foo2=None):
            """doc child"""
            self.foo2 = 42
    cnode = ChildNode()
    assert cnode.train.__doc__ == "doc child"
    cnode.train(X, foo2='abc')
    assert cnode.foo2 == 42
    assert get_signature(cnode.train) == 'self, x, foo2'

def test_signatures_with_doc_in_ancestor():
    # first try on a subclass of Node if
    # the signature and the docstring are exported to
    # the public method
    class AncestorNode(mdp.Node):
        def _train(self, x, foo=None):
            """doc ancestor"""
            self.foo = 42
    anode = AncestorNode()
    assert anode.train.__doc__ == "doc ancestor"
    anode.train(X, foo='abc')
    assert anode.foo == 42
    assert get_signature(anode.train) == 'self, x, foo'

    # now try on a subclass of it
    class ChildNode(AncestorNode):
        def _train(self, x, foo2=None):
            self.foo2 = 42
    cnode = ChildNode()
    assert cnode.train.__doc__ == "doc ancestor"
    cnode.train(X, foo2='abc')
    assert cnode.foo2 == 42
    assert get_signature(cnode.train) == 'self, x, foo2'

def test_signatures_with_doc_in_child():
    # first try on a subclass of Node if
    # the signature and the docstring are exported to
    # the public method
    class AncestorNode(mdp.Node):
        def _train(self, x, foo=None):
            self.foo = 42
    anode = AncestorNode()
    anode.train(X, foo='abc')
    assert anode.foo == 42
    assert get_signature(anode.train) == 'self, x, foo'

    # now try on a subclass of it
    class ChildNode(AncestorNode):
        def _train(self, x, foo2=None):
            """doc child"""
            self.foo2 = 42
    cnode = ChildNode()
    assert cnode.train.__doc__ == "doc child"
    cnode.train(X, foo2='abc')
    assert cnode.foo2 == 42
    assert get_signature(cnode.train) == 'self, x, foo2'



"""A lightweight python templating engine.    Templet version 2 beta.

Slighlty modifed version for MDP (different indentation handling).


Supports two templating idioms:
        1. template functions using @stringfunction and @unicodefunction
        2. template classes inheriting from StringTemplate and UnicodeTemplate

Each template function is marked with the attribute @stringfunction
or @unicodefunction.    Template functions will be rewritten to expand
their document string as a template and return the string result.
For example:

        @stringtemplate
        def myTemplate(animal, thing):
            "the $animal jumped over the $thing."

        print myTemplate('cow', 'moon')

The template language understands the following forms:

        $myvar - inserts the value of the variable 'myvar'
        ${...} - evaluates the expression and inserts the result
        ${{...}} - executes enclosed code; use 'out.append(text)' to insert text
        $$ - an escape for a single $
        $ (at the end of the line) - a line continuation

Template functions are compiled into code that accumulates a list of
strings in a local variable 'out', and then returns the concatenation
of them.    If you want do do complicated computation, you can append
to 'out' directly inside a ${{...}} block.

Another alternative is to use template classes.

Each template class is a subclass of StringTemplate or UnicodeTemplate.
Template classes should define a class attribute 'template' that
contains the template code.    Also, any class attribute ending with
'_template' will be compiled into a template method.

Use a template class by instantiating it with a dictionary or
keyword arguments.    Get the expansion by converting the instance
to a string.    For example:

        class MyTemplate(templet.Template):
            template = "the $animal jumped over the $thing."

        print MyTemplate(animal='cow', thing='moon')

Within a template class, the template language is similar to a template
function, but 'self.write' should be used to build the string inside
${{..}} blocks.    Also, there is a shorthand for calling template methods:

        $<sub_template> - shorthand for '${{self.sub_template(vars())}}'

This idiom is helpful for decomposing a template and when subclassing.

A longer example:

    import cgi
    class RecipeTemplate(templet.Template):
        template = r'''
            <html><head><title>$dish</title></head>
            <body>
            $<header_template>
            $<body_template>
            </body></html>
        '''
        header_template = r'''
            <h1>${cgi.escape(dish)}</h1>
        '''
        body_template = r'''
            <ol>
            ${{
                for item in ingredients:
                    self.write('<li>', item, '\n')
            }}
            </ol>
        '''

This template can be expanded as follows:

    print RecipeTemplate(dish='burger', ingredients=['bun', 'beef', 'lettuce'])

And it can be subclassed like this:

    class RecipeWithPriceTemplate(RecipeTemplate):
        header_template = "<h1>${cgi.escape(dish)} - $$$price</h1>\n"

Templet is by David Bau and was inspired by Tomer Filiba's Templite class.
For details, see http://davidbau.com/templet

Templet is posted by David Bau under BSD-license terms.

Copyright (c) 2007, David Bau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
         this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in the
         documentation and/or other materials provided with the distribution.

    3. Neither the name of Templet nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from __future__ import print_function
from builtins import str
from past.builtins import basestring
from builtins import object

import sys, re, inspect
from future.utils import with_metaclass


class _TemplateBuilder(object):

    __pattern = re.compile(r"""\$(   # Directives begin with a $
                \$                 | # $$ is an escape for $
                [^\S\n]*\n         | # $\n is a line continuation
                [_a-z][_a-z0-9]*   | # $simple Python identifier
                \{(?!\{)[^\}]*\}   | # ${...} expression to eval
                \{\{.*?\}\}        | # ${{...}} multiline code to exec
                <[_a-z][_a-z0-9]*> | # $<sub_template> method call
            )(?:(?:(?<=\}\})|(?<=>))[^\S\n]*\n)?  # eat some trailing newlines
        """, re.IGNORECASE | re.VERBOSE | re.DOTALL)

    def __init__(self, constpat, emitpat, callpat=None):
        self.constpat, self.emitpat, self.callpat = constpat, emitpat, callpat

    def __realign(self, str, spaces=''):
        """Removes any leading empty columns of spaces and an initial
        empty line.

        This is important for embedded Python code.
        """
        lines = str.splitlines()
        if lines and not lines[0].strip(): del lines[0]
        lspace = [len(l) - len(l.lstrip()) for l in lines if l.lstrip()]
        margin = len(lspace) and min(lspace)
        return '\n'.join((spaces + l[margin:]) for l in lines)

    def build(self, template, filename, s=''):
        code = []
        for i, part in enumerate(self.__pattern.split(template)):
            if i % 2 == 0:
                if part:
                    code.append(s + self.constpat % repr(part))
            else:
                if not part or (part.startswith('<') and self.callpat is None):
                    raise SyntaxError('Unescaped $ in ' + filename)
                elif part.endswith('\n'):
                    continue
                elif part == '$':
                    code.append(s + self.emitpat % '"$"')
                elif part.startswith('{{'):
                    code.append(self.__realign(part[2:-2], s))
                elif part.startswith('{'):
                    code.append(s + self.emitpat % part[1:-1])
                elif part.startswith('<'):
                    code.append(s + self.callpat % part[1:-1])
                else:
                    code.append(s + self.emitpat % part)
        return '\n'.join(code)

def _exec(code, globals, locals):
    exec(code, globals, locals)

class _TemplateMetaClass(type):

    __builder = _TemplateBuilder(
        'self.out.append(%s)', 'self.write(%s)', 'self.%s(vars())')

    def __compile(cls, template, n):
        globals = sys.modules[cls.__module__].__dict__
        if '__file__' not in globals: filename = '<%s %s>' % (cls.__name__, n)
        else: filename = '%s: <%s %s>' % (globals['__file__'], cls.__name__, n)
        code = compile(cls.__builder.build(template, filename), filename, 'exec')
        def expand(self, __dict = None, **kw):
            if __dict: kw.update([i for i in __dict.items() if i[0] not in kw])
            kw['self'] = self
            _exec(code, globals, kw)
        return expand

    def __init__(cls, *args):
        for attr, val in list(cls.__dict__.items()):
            if attr == 'template' or attr.endswith('_template'):
                if isinstance(val, basestring):
                    setattr(cls, attr, cls.__compile(val, attr))
        type.__init__(cls, *args)


class StringTemplate(with_metaclass(_TemplateMetaClass, object)):
    """A base class for string template classes."""

    def __init__(self, *args, **kw):
        self.out = []
        self.template(*args, **kw)

    def write(self, *args):
        self.out.extend([str(a) for a in args])

    def __str__(self):
        return ''.join(self.out)


# The original version of templet called StringTemplate "Template"
Template = StringTemplate


class UnicodeTemplate(with_metaclass(_TemplateMetaClass, object)):
    """A base class for unicode template classes."""

    def __init__(self, *args, **kw):
        self.out = []
        self.template(*args, **kw)

    def write(self, *args):
        self.out.extend([str(a) for a in args])

    def __unicode__(self):
        return u''.join(self.out)

    def __str__(self):
        return str(self).encode('utf-8')


def _templatefunction(func, listname, stringtype):
    globals, locals = sys.modules[func.__module__].__dict__, {}
    if '__file__' not in globals: filename = '<%s>' % func.__name__
    else: filename = '%s: <%s>' % (globals['__file__'], func.__name__)
    builder = _TemplateBuilder('%s.append(%%s)' % listname,
                                                         '%s.append(%s(%%s))' % (listname, stringtype))
    args = inspect.getargspec(func)
    code = [
        'def %s%s:' % (func.__name__, inspect.formatargspec(*args)),
        ' %s = []' % listname,
        builder.build(func.__doc__, filename, ' '),
        ' return "".join(%s)' % listname]
    code = compile('\n'.join(code), filename, 'exec')
    exec(code, globals, locals)
    return locals[func.__name__]

def stringfunction(func):
    """Function attribute for string template functions"""
    return _templatefunction(func, listname='out', stringtype='str')

def unicodefunction(func):
    """Function attribute for unicode template functions"""
    return _templatefunction(func, listname='out', stringtype='unicode')


# When executed as a script, run some testing code.
if __name__ == '__main__':
    ok = True
    def expect(actual, expected):
        global ok
        if expected != actual:
            print("error - got:\n%s" % repr(actual))
            ok = False
    class TestAll(Template):
        """A test of all the $ forms"""
        template = r"""
            Bought: $count ${name}s$
             at $$$price.
            ${{
                for i in xrange(count):
                    self.write(TestCalls(vars()), "\n")    # inherit all the local $vars
            }}
            Total: $$${"%.2f" % (count * price)}
        """
    class TestCalls(Template):
        """A recursive test"""
        template = "$name$i ${*[TestCalls(name=name[0], i=n) for n in xrange(i)]}"
    expect(
        str(TestAll(count=5, name="template call", price=1.23)),
        "Bought: 5 template calls at $1.23.\n"
        "template call0 \n"
        "template call1 t0 \n"
        "template call2 t0 t1 t0 \n"
        "template call3 t0 t1 t0 t2 t0 t1 t0 \n"
        "template call4 t0 t1 t0 t2 t0 t1 t0 t3 t0 t1 t0 t2 t0 t1 t0 \n"
        "Total: $6.15\n")
    class TestBase(Template):
        template = r"""
            <head>$<head_template></head>
            <body>$<body_template></body>
        """
    class TestDerived(TestBase):
        head_template = "<title>$name</title>"
        body_template = "${TestAll(vars())}"
    expect(
        str(TestDerived(count=4, name="template call", price=2.88)),
        "<head><title>template call</title></head>\n"
        "<body>"
        "Bought: 4 template calls at $2.88.\n"
        "template call0 \n"
        "template call1 t0 \n"
        "template call2 t0 t1 t0 \n"
        "template call3 t0 t1 t0 t2 t0 t1 t0 \n"
        "Total: $11.52\n"
        "</body>\n")
    class TestUnicode(UnicodeTemplate):
        template = u"""
            \N{Greek Small Letter Pi} = $pi
        """
    expect(
        str(TestUnicode(pi = 3.14)),
        u"\N{Greek Small Letter Pi} = 3.14\n")
    goterror = False
    try:
        class TestError(Template):
            template = 'Cost of an error: $0'
    except SyntaxError:
        goterror = True
    if not goterror:
        print('TestError failed')
        ok = False
    @stringfunction
    def testBasic(name):
        "Hello $name."
    expect(testBasic('Henry'), "Hello Henry.")
    @stringfunction
    def testReps(a, count=5): r"""
        ${{ if count == 0: return '' }}
        $a${testReps(a, count - 1)}"""
    expect(
        testReps('foo'),
        "foofoofoofoofoo")
    @unicodefunction
    def testUnicode(count=4): u"""
        ${{ if not count: return '' }}
        \N{BLACK STAR}${testUnicode(count - 1)}"""
    expect(
        testUnicode(count=10),
        u"\N{BLACK STAR}" * 10)
    if ok: print("OK")
